from dataclasses import dataclass
from typing import Tuple

from lark.visitors import Interpreter


@dataclass(frozen=True)
class Let:
    name: str
    value: Tuple[str, ...]


class DefinitionInterpreter(Interpreter):
    def interpret(self, tree):
        children = self.visit_children(tree)
        assert len(children) == 2
        return Let(children[0].value, tuple(children[1]))

    def value(self, tree):
        children = self.visit_children(tree)
        assert len(children) == 1
        value = children[0].value
        assert isinstance(value, str)
        return value.upper()



import pytest

from pharmpy.deps import sympy
from pharmpy.model import Assignment
from pharmpy.model.external.nonmem.nmtran_parser import NMTranParser
from pharmpy.model.external.nonmem.parsing import parse_table_columns
from pharmpy.modeling import read_model_from_string


def symbol(x):
    return sympy.Symbol(x)


def test_simple_parse():
    parser = NMTranParser()

    model = parser.parse('$PROBLEM MYPROB\n')

    assert len(model.records) == 1
    assert type(model.records[0]).__name__ == 'ProblemRecord'
    assert str(model) == '$PROBLEM MYPROB\n'

    model2_str = ';Comment\n   $PROBLEM     TW2\n'
    model2 = parser.parse(model2_str)

    assert len(model2.records) == 2
    assert type(model2.records[0]).__name__ == 'RawRecord'
    assert type(model2.records[1]).__name__ == 'ProblemRecord'

    assert str(model2) == model2_str


def test_round_trip(pheno_path):
    parser = NMTranParser()

    with open(pheno_path, 'r') as fh:
        content = fh.read()
    model = parser.parse(content)
    assert str(model) == content


@pytest.mark.parametrize(
    'buf,columns',
    [
        ('$TABLE ID TIME', [['ID', 'TIME', 'DV', 'PRED', 'RES', 'WRES']]),
        ('$TABLE ID TIME NOAPPEND', [['ID', 'TIME']]),
        (
            '$TABLE ID TIME CIPREDI=CONC NOAPPEND',
            [['ID', 'TIME', 'CIPREDI']],
        ),
        (
            '$TABLE ID TIME CONC=CIPREDI NOAPPEND',
            [['ID', 'TIME', 'CIPREDI']],
        ),
        (
            '$TABLE ID TIME CONC=CIPREDI NOAPPEND\n$TABLE ID TIME CONC',
            [['ID', 'TIME', 'CIPREDI'], ['ID', 'TIME', 'CIPREDI', 'DV', 'PRED', 'RES', 'WRES']],
        ),
    ],
)
def test_table_columns(buf, columns):
    parser = NMTranParser()

    cs = parser.parse(f'$PROBLEM\n{buf}')
    parsed_columns = parse_table_columns(cs, netas=2)
    assert parsed_columns == columns


def test_parse_multiple_dvs():
    code1 = """$PROBLEM PHENOBARB PD
$INPUT ID TIME DV AMT DVID
$DATA pheno_pd.csv IGNORE=@
$SUBROUTINES ADVAN1 TRANS2
$ABBR REPLACE ETA_E0=ETA(3)
$ABBR REPLACE ETA_S=ETA(4)
$PK
S = THETA(4)*EXP(ETA_S)
E0 = THETA(3)*EXP(ETA_E0)
VC = THETA(1) * EXP(ETA(1))
CL = THETA(2) * EXP(ETA(2))
V=VC
$ERROR
C = A(1)/V
Y = C*(1+EPS(1))
E = A(1)*S/V + E0
Y_2 = E + E*EPS(2)
IF (DVID.EQ.1) THEN
    Y = Y
ELSE
    Y = Y_2
END IF
$ESTIMATION MAXEVAL=9999 SIGDIGITS=4 POSTHOC
$THETA (0, 95.0648) FIX ; POP_VC
$THETA (0, 10.3416) FIX ; POP_CL
$THETA  (0,0.1) ; POP_E0
$THETA  (0,0.1) ; POP_S
$OMEGA BLOCK(2) FIX
0.0605492	; IIV_VC
0.000294284	; IIV_VC_IIV_CL
1.13303e-05	; IIV_CL
$OMEGA  0.09 ; IIV_E0
$OMEGA  0.09 ; IIV_S
$SIGMA 1.16066 FIX; PK RUV_PROP
$SIGMA  0.09 ; sigma"""

    model = read_model_from_string(code1)
    assert Assignment(symbol('Y_1'), symbol('Y')) not in model.statements.after_odes
    assert Assignment(symbol('Y_2'), symbol('Y_2')) not in model.statements.after_odes
    assert (
        Assignment(symbol('Y'), symbol('C') * (1 + symbol('EPS_1'))) in model.statements.after_odes
    )
    assert (
        Assignment(symbol('Y_2'), symbol('E') + symbol('E') * symbol('EPS_2'))
        in model.statements.after_odes
    )
    assert model.dependent_variables == {symbol('Y'): 1, symbol('Y_2'): 2}
    assert model.observation_transformation == {
        symbol('Y'): symbol('Y'),
        symbol('Y_2'): symbol('Y_2'),
    }

    code2 = """$PROBLEM PHENOBARB PD
$INPUT ID TIME DV AMT DVID
$DATA pheno_pd.csv IGNORE=@
$SUBROUTINES ADVAN1 TRANS2
$ABBR REPLACE ETA_E0=ETA(3)
$ABBR REPLACE ETA_S=ETA(4)
$PK
S = THETA(4)*EXP(ETA_S)
E0 = THETA(3)*EXP(ETA_E0)
VC = THETA(1) * EXP(ETA(1))
CL = THETA(2) * EXP(ETA(2))
V=VC
$ERROR
C = A(1)/V
E = A(1)*S/V + E0
IF (DVID.EQ.1) THEN
    Y = C*(1+EPS(1))
ELSE
    Y = E + E*EPS(2)
END IF
$ESTIMATION MAXEVAL=9999 SIGDIGITS=4 POSTHOC
$THETA (0, 95.0648) FIX ; POP_VC
$THETA (0, 10.3416) FIX ; POP_CL
$THETA  (0,0.1) ; POP_E0
$THETA  (0,0.1) ; POP_S
$OMEGA BLOCK(2) FIX
0.0605492	; IIV_VC
0.000294284	; IIV_VC_IIV_CL
1.13303e-05	; IIV_CL
$OMEGA  0.09 ; IIV_E0
$OMEGA  0.09 ; IIV_S
$SIGMA 1.16066 FIX; PK RUV_PROP
$SIGMA  0.09 ; sigma"""

    model2 = read_model_from_string(code2)
    assert (
        Assignment(symbol('Y_1'), symbol('C') * (1 + symbol('EPS_1')))
        in model2.statements.after_odes
    )
    assert (
        Assignment(symbol('Y_2'), symbol('E') + symbol('E') * symbol('EPS_2'))
        in model2.statements.after_odes
    )
    assert model2.dependent_variables == {symbol('Y_1'): 1, symbol('Y_2'): 2}
    assert model2.observation_transformation == {
        symbol('Y_1'): symbol('Y_1'),
        symbol('Y_2'): symbol('Y_2'),
    }



from sympy.physics import units

from pharmpy.modeling import get_unit_of


def test_get_unit_of(load_model_for_test, testdata):
    model = load_model_for_test(testdata / "nonmem" / "pheno_real.mod")
    assert get_unit_of(model, "Y") == units.milligram / units.l
    assert get_unit_of(model, "V") == units.l
    assert get_unit_of(model, "WGT") == units.kilogram



from typing import Iterable, Literal, Optional, Tuple, Union

from pharmpy.model import Model
from pharmpy.workflows import ModelEntry, Task, Workflow, WorkflowBuilder

SupportedExternalTools = Literal['nonmem', 'nlmixr', 'rxode']


def create_workflow(
    model_or_models: Optional[Union[Model, Iterable[Model]]] = None,
    n: Optional[int] = None,
    tool: Optional[SupportedExternalTools] = None,
) -> Workflow[Union[Model, Tuple[Model, ...]]]:
    """Run modelfit tool.

    Parameters
    ----------
    model_or_models : Model
        A list of models are one single model object
    n : int
        Number of models to fit. This is only used if the tool is going to be combined with other tools.
    tool : str
        Which tool to use for fitting. Currently, 'nonmem', 'nlmixr', 'rxode' can be used.

    Returns
    -------
    ModelfitResults
        Modelfit tool result object

    Examples
    --------
    >>> from pharmpy.modeling import load_example_model
    >>> model = load_example_model("pheno")
    >>> from pharmpy.tools import run_modelfit     # doctest: +SKIP
    >>> run_modelfit(model)   # doctest: +SKIP
    """
    if model_or_models:
        if not isinstance(model_or_models, Iterable):
            model_or_models = [model_or_models]
        modelentries = [ModelEntry.create(model=model) for model in model_or_models]
    wf = create_fit_workflow(modelentries, n, tool)
    wf = wf.replace(name="modelfit")
    if len(modelentries) == 1 or (modelentries is None and n is None):
        post_process_results = post_process_results_one
    else:
        post_process_results = post_process_results_many
    task_result: Task[Union[Model, Tuple[Model, ...]]] = Task('results', post_process_results)
    wb = WorkflowBuilder(wf)
    wb.add_task(task_result, predecessors=wf.output_tasks)
    wf = Workflow(wb)
    return wf


def create_fit_workflow(modelentries=None, n=None, tool=None):
    execute_model = retrieve_from_database_or_execute_model_with_tool(tool)

    wb = WorkflowBuilder()
    if modelentries is None:
        if n is None:
            task = Task('run', execute_model)
            wb.add_task(task)
        else:
            for i in range(n):
                task = Task(f'run{i}', execute_model)
                wb.add_task(task)
    elif isinstance(modelentries, ModelEntry):
        task = Task('run', execute_model, modelentries)
        wb.add_task(task)
    else:
        assert all(isinstance(m, ModelEntry) for m in modelentries)
        for i, model_entry in enumerate(modelentries):
            task = Task(f'run{i}', execute_model, model_entry)
            wb.add_task(task)
    return Workflow(wb)


def post_process_results_one(context, *model_entry: ModelEntry):
    return model_entry[0].modelfit_results


def post_process_results_many(context, *modelentries: ModelEntry):
    return tuple([m.modelfit_results for m in modelentries])


def retrieve_from_database_or_execute_model_with_tool(tool):
    def task(context, model_entry):
        assert isinstance(model_entry, ModelEntry)
        model = model_entry.model
        try:
            db_model_entry = context.model_database.retrieve_model_entry(model.name)
        except (KeyError, AttributeError, FileNotFoundError):
            db_model_entry = None

        if db_model_entry and db_model_entry.modelfit_results is not None:
            if model.has_same_dataset_as(db_model_entry.model):
                return model_entry.attach_results(
                    db_model_entry.modelfit_results, db_model_entry.log
                )

        # NOTE: Fallback to executing the model
        execute_model = get_execute_model(tool)
        return execute_model(model_entry, context)

    return task


def get_execute_model(tool: Optional[SupportedExternalTools]):
    from pharmpy.tools.modelfit import conf

    if tool is None:
        tool = conf.default_tool

    if tool == 'nonmem':
        from pharmpy.tools.external.nonmem.run import execute_model
    elif tool == 'nlmixr':
        from pharmpy.tools.external.nlmixr.run import execute_model
    elif tool == 'rxode':
        from pharmpy.tools.external.rxode.run import execute_model
    else:
        raise ValueError(f"Unknown estimation tool {tool}")

    return execute_model



import math
import re
from typing import cast

from pharmpy.deps import numpy as np
from pharmpy.internals.math import flattened_to_symmetric, triangular_root
from pharmpy.internals.parse import AttrToken, AttrTree
from pharmpy.internals.parse.generic import (
    eval_token,
    insert_after,
    insert_before_or_at_end,
    remove_token_and_space,
)
from pharmpy.model import ModelSyntaxError

from .record import Record


class OmegaRecord(Record):
    def parse(self):
        """Parse the omega record

        Return a list with one tuple for each block. Each block will have a list of names,
        a list of inits, fixedness and same.
        """
        block = self.root.find('block')
        bare_block = self.root.find('bare_block')
        same = bool(self.root.find('same'))
        blocks = []

        if not (block or bare_block):
            for node in self.root.subtrees('diag_item'):
                init = cast(float, eval_token(node.subtree('init').leaf('NUMERIC')))
                fixed = bool(node.find('FIX'))
                sd = bool(node.find('SD'))
                var = bool(node.find('VAR'))
                n = cast(int, eval_token(node.subtree('n').leaf('INT'))) if node.find('n') else 1
                if sd and var:
                    name = '(anonymous)' if self.name is None else self.name.upper()
                    raise ModelSyntaxError(
                        f'Initial estimate for {name} cannot be both'
                        f' on SD and VAR scale\n{self.root}'
                    )
                if init == 0 and not fixed:
                    name = '(anonymous)' if self.name is None else self.name.upper()
                    raise ModelSyntaxError(
                        f'If initial estimate for {name} is 0 it must be set to FIX'
                    )
                if sd:
                    init = init**2
                name = self._get_name(node)
                for _ in range(n):
                    block = ([name], [init], fixed, False)
                    blocks.append(block)
        else:
            inits = []
            fix, sd, corr, cholesky = self._block_flags()
            comments = []
            for node in self.root.subtrees('omega'):
                init = cast(float, eval_token(node.subtree('init').leaf('NUMERIC')))
                n = cast(int, eval_token(node.subtree('n').leaf('INT'))) if node.find('n') else 1
                inits += [init] * n
                comment = self._get_name(node)
                comments.append(comment)
                if n > 1:
                    comments.extend([None] * (n - 1))
            if not same:
                size = cast(int, eval_token(self.root.subtree('block').subtree('size').leaf('INT')))
                if size != triangular_root(len(inits)):
                    raise ModelSyntaxError('Wrong number of inits in BLOCK')
                if not cholesky:
                    A = flattened_to_symmetric(inits)
                    if corr:
                        for i in range(size):
                            for j in range(size):
                                if i != j:
                                    if sd:
                                        A[i, j] = A[i, i] * A[j, j] * A[i, j]
                                    else:
                                        A[i, j] = math.sqrt(A[i, i]) * math.sqrt(A[j, j]) * A[i, j]
                    if sd:
                        np.fill_diagonal(A, A.diagonal() ** 2)
                else:
                    L = np.zeros((size, size))
                    inds = np.tril_indices_from(L)
                    L[inds] = inits
                    A = L @ L.T
                label_index = 0
                names = []
                inits = []
                for i in range(size):
                    for j in range(0, i + 1):
                        comment = comments[label_index]
                        init = A[i, j]
                        names.append(comment)
                        inits.append(init)
                        label_index += 1
                block = (names, inits, fix, False)
            else:
                block = (None, None, None, True)
            blocks.append(block)
        return blocks

    def _get_name(self, node):
        name = None
        found = False
        for subnode in self.root.tree_walk():
            if id(subnode) == id(node):
                found = True
                continue
            if found and (subnode.rule == 'omega' or subnode.rule == 'diag_item'):
                break
            if found and (subnode.rule == 'NEWLINE' or subnode.rule == 'COMMENT'):
                m = re.search(r';\s*([a-zA-Z_]\w*)', str(subnode))
                if m:
                    name = m.group(1)
                    break
        return name

    def _block_flags(self):
        """Get a tuple of all interesting flags for block"""
        fix = bool(self.root.find('FIX'))
        var = bool(self.root.find('VAR'))
        sd = bool(self.root.find('SD'))
        cov = bool(self.root.find('COV'))
        corr = bool(self.root.find('CORR'))
        cholesky = bool(self.root.find('CHOLESKY'))
        for node in self.root.subtrees('omega'):
            if node.find('FIX'):
                if fix:
                    raise ModelSyntaxError('Cannot specify option FIX more than once')
                else:
                    fix = True
            if node.find('VAR'):
                if var or sd or cholesky:
                    raise ModelSyntaxError(
                        'Cannot specify either option VARIANCE, SD or ' 'CHOLESKY more than once'
                    )
                else:
                    var = True
            if node.find('SD'):
                if sd or var or cholesky:
                    raise ModelSyntaxError(
                        'Cannot specify either option VARIANCE, SD or ' 'CHOLESKY more than once'
                    )
                else:
                    sd = True
            if node.find('COV'):
                if cov or corr:
                    raise ModelSyntaxError(
                        'Cannot specify either option COVARIANCE or ' 'CORRELATION more than once'
                    )
                else:
                    cov = True
            if node.find('CORR'):
                if corr or cov:
                    raise ModelSyntaxError(
                        'Cannot specify either option COVARIANCE or ' 'CORRELATION more than once'
                    )
                else:
                    corr = True
            if node.find('CHOLESKY'):
                if cholesky or var or sd:
                    raise ModelSyntaxError(
                        'Cannot specify either option VARIANCE, SD or ' 'CHOLESKY more than once'
                    )
                else:
                    cholesky = True
        return fix, sd, corr, cholesky

    def update(self, parameters):
        """From a Parameters update the OMEGAs in this record"""
        block = self.root.find('block')
        bare_block = self.root.find('bare_block')
        if not (block or bare_block):
            i = 0
            size = 1
            new_nodes = []
            for node in self.root.children:
                if not isinstance(node, AttrTree) or node.rule != 'diag_item':
                    new_nodes.append(node)
                else:
                    sd = bool(node.find('SD'))
                    fix = bool(node.find('FIX'))
                    n = (
                        cast(int, eval_token(node.subtree('n').leaf('INT')))
                        if node.find('n')
                        else 1
                    )
                    new_inits = []
                    new_fix = []
                    for j in range(i, i + n):
                        if not sd:
                            value = parameters[j].init
                        else:
                            value = parameters[j].init ** 0.5
                        new_inits.append(value)
                        new_fix.append(parameters[j].fix)
                    new_init = new_inits[0]
                    if n == 1 or (
                        new_inits.count(new_init) == len(new_inits)
                        and new_fix.count(new_fix[0]) == len(new_fix)
                    ):  # All equal?
                        init = node.subtree('init')
                        if eval_token(init.leaf('NUMERIC')) != new_init:
                            new_init = int(new_init) if float(new_init).is_integer() else new_init
                            init = init.replace_first(AttrToken('NUMERIC', str(new_init)))
                        node = node.replace_first(init)
                        if new_fix[0] != fix:
                            if new_fix[0]:
                                node = insert_before_or_at_end(
                                    node, 'RPAR', [AttrToken('WS', ' '), AttrToken('FIX', 'FIX')]
                                )
                            else:
                                node = remove_token_and_space(node, 'FIX')
                        new_nodes.append(node)
                    else:
                        # Need to split xn
                        node = node.remove('n')
                        for j, new_init in enumerate(new_inits):
                            init = node.subtree('init')
                            if eval_token(init.leaf('NUMERIC')) != new_init:
                                new_init = (
                                    int(new_init) if float(new_init).is_integer() else new_init
                                )
                                init = init.replace_first(AttrToken('NUMERIC', str(new_init)))
                            node = node.replace_first(init)
                            if new_fix[j] != fix:
                                node = insert_before_or_at_end(
                                    node, 'RPAR', [AttrToken('WS', ' '), AttrToken('FIX', 'FIX')]
                                )
                            else:
                                node = remove_token_and_space(node, 'FIX')
                            new_nodes.append(node)
                            if j != len(new_inits) - 1:  # Not the last
                                new_nodes.append(AttrTree.create('ws', {'WS': ' '}))
                    i += n
            tree = AttrTree(self.root.rule, tuple(new_nodes))
            return OmegaRecord(self.name, self.raw_name, tree)
        else:
            same = bool(self.root.find('same'))
            if same:
                return self
            size = cast(int, eval_token(self.root.subtree('block').subtree('size').leaf('INT')))
            fix, sd, corr, cholesky = self._block_flags()
            inits = [param.init for param in parameters]
            new_fix = [param.fix for param in parameters]
            if len(set(new_fix)) != 1:  # Not all true or all false
                raise ValueError('Cannot only fix some parameters in block')

            A = flattened_to_symmetric(inits)

            if corr:
                for i in range(size):
                    for j in range(size):
                        if i != j:
                            A[i, j] = A[i, j] / (math.sqrt(A[i, i]) * math.sqrt(A[j, j]))
            if sd:
                np.fill_diagonal(A, A.diagonal() ** 0.5)

            if cholesky:
                A = np.linalg.cholesky(A)

            inds = np.tril_indices_from(A)
            array = list(A[inds])
            i = 0
            new_nodes = []
            for node in self.root.children:
                if not isinstance(node, AttrTree) or node.rule != 'omega':
                    new_nodes.append(node)
                else:
                    n = (
                        cast(int, eval_token(node.subtree('n').leaf('INT')))
                        if node.find('n')
                        else 1
                    )
                    if array[i : i + n].count(array[i]) == n:  # All equal?
                        init = node.subtree('init')
                        new_init = array[i]
                        if eval_token(init.leaf('NUMERIC')) != new_init:
                            new_init = int(new_init) if float(new_init).is_integer() else new_init
                            init = init.replace_first(AttrToken('NUMERIC', str(new_init)))
                        node = node.replace_first(init)
                        new_nodes.append(node)
                    else:
                        # NOTE: Split xn
                        new_children = []
                        for child in node.children:
                            if child.rule not in ('n', 'LPAR', 'RPAR'):
                                new_children.append(child)
                        node = AttrTree(node.rule, tuple(new_children))
                        for j, new_init in enumerate(array[i : i + n]):
                            init = node.subtree('init')
                            if eval_token(init.leaf('NUMERIC')) != new_init:
                                new_init = (
                                    int(new_init) if float(new_init).is_integer() else new_init
                                )
                                init = init.replace_first(AttrToken('NUMERIC', str(new_init)))
                            node = node.replace_first(init)
                            new_nodes.append(node)
                            if j != n - 1:  # NOTE: Not the last
                                new_nodes.append(AttrTree.create('ws', {'WS': ' '}))
                    i += n
            tree = AttrTree(self.root.rule, tuple(new_nodes))
            if new_fix[0] != fix:
                if new_fix[0]:
                    tree = insert_after(
                        tree, 'block', [AttrToken('WS', ' '), AttrToken('FIX', 'FIX')]
                    )
                else:
                    tree = remove_token_and_space(tree, 'FIX', recursive=True)
            return OmegaRecord(self.name, self.raw_name, tree)

    def remove(self, inds):
        """Remove some etas from block given eta names
        inds - list of tuples of block index and eta index within block
        """
        if len(inds) == 0:
            return self

        block = self.root.find('block')
        bare_block = self.root.find('bare_block')
        same = bool(self.root.find('same'))
        if not (block or bare_block):
            keep = []
            i = 0
            eta_inds = {ind for ind, _ in inds}
            in_keep = True
            for node in self.root.children:
                if node.rule == 'diagonal':
                    in_keep = False
                if node.rule == 'diag_item':
                    in_keep = i not in eta_inds
                    i += 1
                if in_keep:
                    keep.append(node)
            tree = AttrTree(self.root.rule, tuple(keep))
            return OmegaRecord(self.name, self.raw_name, tree)
        elif same and not bare_block:
            indices = {ind for _, ind in inds}
            tree = self.root.replace_first(
                (block := self.root.subtree('block')).replace_first(
                    block.subtree('size').replace_first(
                        AttrToken('INT', str(len(self) - len(indices)))
                    )
                )
            )
            return OmegaRecord(self.name, self.raw_name, tree)
        elif block:
            fix, sd, corr, cholesky = self._block_flags()
            inits = []
            indices = {ind for _, ind in inds}
            for node in self.root.subtrees('omega'):
                init = cast(float, eval_token(node.subtree('init').leaf('NUMERIC')))
                n = cast(int, eval_token(node.subtree('n').leaf('INT'))) if node.find('n') else 1
                inits += [init] * n
            A = flattened_to_symmetric(inits)
            A = np.delete(A, list(indices), axis=0)
            A = np.delete(A, list(indices), axis=1)
            s = f' BLOCK({len(A)})'
            if fix:
                s += ' FIX'
            if sd:
                s += ' SD'
            if corr:
                s += ' CORR'
            if cholesky:
                s += ' CHOLESKY'
            s += '\n'
            for row in range(0, len(A)):
                s += ' '.join(np.atleast_1d(A[row, 0 : (row + 1)]).astype(str)) + '\n'

            temp = self.raw_name + s
            from .factory import create_record

            return create_record(temp)  # FIXME: No need to reparse

    def __len__(self):
        # Number of blocks in record
        block = self.root.find('block')
        bare_block = self.root.find('bare_block')
        if not (block or bare_block):
            size = 0
            for node in self.root.subtrees('diag_item'):
                n = cast(int, eval_token(node.subtree('n').leaf('INT'))) if node.find('n') else 1
                size += n
        else:
            size = 1
        return size



from typing import List

from lark.visitors import Interpreter

from .statement.definition import DefinitionInterpreter
from .statement.feature.absorption import AbsorptionInterpreter
from .statement.feature.covariate import CovariateInterpreter
from .statement.feature.direct_effect import DirectEffectInterpreter
from .statement.feature.effect_comp import EffectCompInterpreter
from .statement.feature.elimination import EliminationInterpreter
from .statement.feature.indirect_effect import IndirectEffectInterpreter
from .statement.feature.lagtime import LagTimeInterpreter
from .statement.feature.metabolite import MetaboliteInterpreter
from .statement.feature.peripherals import PeripheralsInterpreter
from .statement.feature.transits import TransitsInterpreter
from .statement.statement import Statement


class MFLInterpreter(Interpreter):
    def interpret(self, tree) -> List[Statement]:
        return self.visit_children(tree)

    def definition(self, tree):
        return DefinitionInterpreter().interpret(tree)

    def absorption(self, tree):
        return AbsorptionInterpreter().interpret(tree)

    def elimination(self, tree):
        return EliminationInterpreter().interpret(tree)

    def transits(self, tree):
        return TransitsInterpreter().interpret(tree)

    def peripherals(self, tree):
        return PeripheralsInterpreter().interpret(tree)

    def lagtime(self, tree):
        return LagTimeInterpreter().interpret(tree)

    def covariate(self, tree):
        return CovariateInterpreter().interpret(tree)

    def direct_effect(self, tree):
        return DirectEffectInterpreter().interpret(tree)

    def effect_comp(self, tree):
        return EffectCompInterpreter().interpret(tree)

    def indirect_effect(self, tree):
        return IndirectEffectInterpreter().interpret(tree)

    def metabolite(self, tree):
        return MetaboliteInterpreter().interpret(tree)



import re

import pytest

from pharmpy.deps import numpy as np
from pharmpy.deps import sympy
from pharmpy.modeling.covariate_effect import (
    CovariateEffect,
    _choose_param_inits,
    add_covariate_effect,
    has_covariate_effect,
    remove_covariate_effect,
)

from ..lib import diff


def S(x: str):
    return sympy.Symbol(x)


def test_nan_add_covariate_effect(load_model_for_test, pheno_path):
    model = load_model_for_test(pheno_path)
    data = model.dataset.copy()

    new_col = [np.nan] * 10 + ([1.0] * (len(data.index) - 10))

    data['new_col'] = new_col
    model = model.replace(dataset=data)

    model = add_covariate_effect(model, 'CL', 'new_col', 'cat')

    assert not re.search('NaN', model.model_code)
    assert re.search(r'NEW_COL\.EQ\.-99', model.model_code)


def test_nested_add_covariate_effect(load_model_for_test, testdata):
    model_path = testdata / 'nonmem' / 'pheno.mod'
    model = load_model_for_test(model_path)

    model = add_covariate_effect(model, 'CL', 'WGT', 'exp')
    assert has_covariate_effect(model, 'CL', 'WGT')

    with pytest.warns(UserWarning, match='Covariate effect of WGT on CL already exists'):
        add_covariate_effect(model, 'CL', 'WGT', 'exp')

    assert has_covariate_effect(model, 'CL', 'WGT')

    model = load_model_for_test(model_path)

    model = add_covariate_effect(model, 'CL', 'WGT', 'exp')
    assert has_covariate_effect(model, 'CL', 'WGT')
    assert not has_covariate_effect(model, 'CL', 'APGR')
    model = add_covariate_effect(model, 'CL', 'APGR', 'exp')
    assert has_covariate_effect(model, 'CL', 'WGT')
    assert has_covariate_effect(model, 'CL', 'APGR')

    assert 'CL = CL*CLAPGR*CLWGT' in model.model_code
    assert 'CL = CL*CLWGT' not in model.model_code


@pytest.mark.parametrize(
    ('model_path', 'effects', 'expected', 'allow_nested'),
    [
        (
            ('nonmem', 'pheno.mod'),
            [('CL', 'WGT', 'exp', '*')],
            '@@ -1,0 +2 @@\n'
            '+WGT_MEDIAN = 1.30000\n'
            '@@ -2,0 +4,2 @@\n'
            '+CLWGT = EXP(THETA(3)*(WGT - WGT_MEDIAN))\n'
            '+CL = CL*CLWGT\n',
            False,
        ),
        (
            ('nonmem', 'pheno.mod'),
            [('CL', 'WGT', 'exp', '+')],
            '@@ -1,0 +2 @@\n'
            '+WGT_MEDIAN = 1.30000\n'
            '@@ -2,0 +4,2 @@\n'
            '+CLWGT = EXP(THETA(3)*(WGT - WGT_MEDIAN))\n'
            '+CL = CL + CLWGT\n',
            False,
        ),
        (
            ('nonmem', 'pheno.mod'),
            [('CL', 'APGR', 'cat', '*')],
            '@@ -2,0 +3,22 @@\n'
            '+IF (APGR.EQ.7.0) THEN\n'
            '+    CLAPGR = 1\n'
            '+ELSE IF (APGR.EQ.1.0) THEN\n'
            '+    CLAPGR = THETA(3) + 1\n'
            '+ELSE IF (APGR.EQ.2.0) THEN\n'
            '+    CLAPGR = THETA(4) + 1\n'
            '+ELSE IF (APGR.EQ.3.0) THEN\n'
            '+    CLAPGR = THETA(5) + 1\n'
            '+ELSE IF (APGR.EQ.4.0) THEN\n'
            '+    CLAPGR = THETA(6) + 1\n'
            '+ELSE IF (APGR.EQ.5.0) THEN\n'
            '+    CLAPGR = THETA(7) + 1\n'
            '+ELSE IF (APGR.EQ.6.0) THEN\n'
            '+    CLAPGR = THETA(8) + 1\n'
            '+ELSE IF (APGR.EQ.8.0) THEN\n'
            '+    CLAPGR = THETA(9) + 1\n'
            '+ELSE IF (APGR.EQ.9.0) THEN\n'
            '+    CLAPGR = THETA(10) + 1\n'
            '+ELSE IF (APGR.EQ.10.0) THEN\n'
            '+    CLAPGR = THETA(11) + 1\n'
            '+END IF\n'
            '+CL = CL*CLAPGR\n',
            False,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'WGT', 'exp', '*')],
            '@@ -1,0 +2 @@\n'
            '+WGT_MEDIAN = 1.30000\n'
            '@@ -7,0 +9,2 @@\n'
            '+CLWGT = EXP(THETA(4)*(WGT - WGT_MEDIAN))\n'
            '+CL = CL*CLWGT\n',
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'WGT', 'exp', '+')],
            '@@ -1,0 +2 @@\n'
            '+WGT_MEDIAN = 1.30000\n'
            '@@ -7,0 +9,2 @@\n'
            '+CLWGT = EXP(THETA(4)*(WGT - WGT_MEDIAN))\n'
            '+CL = CL + CLWGT\n',
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'WGT', 'pow', '*')],
            '@@ -1,0 +2 @@\n'
            '+WGT_MEDIAN = 1.30000\n'
            '@@ -7,0 +9,2 @@\n'
            '+CLWGT = (WGT/WGT_MEDIAN)**THETA(4)\n'
            '+CL = CL*CLWGT\n',
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'WGT', 'lin', '*')],
            '@@ -1,0 +2 @@\n'
            '+WGT_MEDIAN = 1.30000\n'
            '@@ -7,0 +9,2 @@\n'
            '+CLWGT = THETA(4)*(WGT - WGT_MEDIAN) + 1\n'
            '+CL = CL*CLWGT\n',
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'APGR', 'cat', '*')],
            '@@ -7,0 +8,22 @@\n'
            '+IF (APGR.EQ.7.0) THEN\n'
            '+    CLAPGR = 1\n'
            '+ELSE IF (APGR.EQ.1.0) THEN\n'
            '+    CLAPGR = THETA(4) + 1\n'
            '+ELSE IF (APGR.EQ.2.0) THEN\n'
            '+    CLAPGR = THETA(5) + 1\n'
            '+ELSE IF (APGR.EQ.3.0) THEN\n'
            '+    CLAPGR = THETA(6) + 1\n'
            '+ELSE IF (APGR.EQ.4.0) THEN\n'
            '+    CLAPGR = THETA(7) + 1\n'
            '+ELSE IF (APGR.EQ.5.0) THEN\n'
            '+    CLAPGR = THETA(8) + 1\n'
            '+ELSE IF (APGR.EQ.6.0) THEN\n'
            '+    CLAPGR = THETA(9) + 1\n'
            '+ELSE IF (APGR.EQ.8.0) THEN\n'
            '+    CLAPGR = THETA(10) + 1\n'
            '+ELSE IF (APGR.EQ.9.0) THEN\n'
            '+    CLAPGR = THETA(11) + 1\n'
            '+ELSE IF (APGR.EQ.10.0) THEN\n'
            '+    CLAPGR = THETA(12) + 1\n'
            '+END IF\n'
            '+CL = CL*CLAPGR\n',
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'WGT', 'piece_lin', '*')],
            '@@ -1,0 +2 @@\n'
            '+WGT_MEDIAN = 1.30000\n'
            '@@ -7,0 +9,6 @@\n'
            '+IF (WGT.LE.WGT_MEDIAN) THEN\n'
            '+    CLWGT = THETA(4)*(WGT - WGT_MEDIAN) + 1\n'
            '+ELSE\n'
            '+    CLWGT = THETA(5)*(WGT - WGT_MEDIAN) + 1\n'
            '+END IF\n'
            '+CL = CL*CLWGT\n',
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'WGT', 'theta - cov + median', '*')],
            '@@ -1,0 +2 @@\n'
            '+WGT_MEDIAN = 1.30000\n'
            '@@ -7,0 +9,2 @@\n'
            '+CLWGT = THETA(4) - WGT + WGT_MEDIAN\n'
            '+CL = CL*CLWGT\n',
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'WGT', 'theta - cov + std', '*')],
            '@@ -1,0 +2 @@\n'
            '+WGT_STD = 0.704565\n'
            '@@ -7,0 +9,2 @@\n'
            '+CLWGT = THETA(4) - WGT + WGT_STD\n'
            '+CL = CL*CLWGT\n',
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'WGT', 'theta1 * (cov/median)**theta2', '*')],
            '@@ -1,0 +2 @@\n'
            '+WGT_MEDIAN = 1.30000\n'
            '@@ -7,0 +9,2 @@\n'
            '+CLWGT = THETA(4)*(WGT/WGT_MEDIAN)**THETA(5)\n'
            '+CL = CL*CLWGT\n',
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'WGT', '((cov/std) - median) * theta', '*')],
            '@@ -1,0 +2,2 @@\n'
            '+WGT_MEDIAN = 1.30000\n'
            '+WGT_STD = 0.704565\n'
            '@@ -7,0 +10,2 @@\n'
            '+CLWGT = THETA(4)*(WGT/WGT_STD - WGT_MEDIAN)\n'
            '+CL = CL*CLWGT\n',
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [
                ('CL', 'WGT', 'exp', '+'),
                ('V', 'WGT', 'exp', '+'),
            ],
            '@@ -1,0 +2 @@\n'
            '+WGT_MEDIAN = 1.30000\n'
            '@@ -7,0 +9,2 @@\n'
            '+CLWGT = EXP(THETA(4)*(WGT - WGT_MEDIAN))\n'
            '+CL = CL + CLWGT\n'
            '@@ -8,0 +12,2 @@\n'
            '+VWGT = EXP(THETA(5)*(WGT - WGT_MEDIAN))\n'
            '+V = V + VWGT\n',
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [
                ('CL', 'WGT', 'exp', '*'),
                ('V', 'WGT', 'exp', '*'),
            ],
            '@@ -1,0 +2 @@\n'
            '+WGT_MEDIAN = 1.30000\n'
            '@@ -7,0 +9,2 @@\n'
            '+CLWGT = EXP(THETA(4)*(WGT - WGT_MEDIAN))\n'
            '+CL = CL*CLWGT\n'
            '@@ -8,0 +12,2 @@\n'
            '+VWGT = EXP(THETA(5)*(WGT - WGT_MEDIAN))\n'
            '+V = V*VWGT\n',
            True,
        ),
        (
            ('nonmem', 'models', 'mox2.mod'),
            [
                ('V', 'WT', 'exp', '*'),
                ('V', 'AGE', 'exp', '*'),
            ],
            '@@ -1,0 +2,2 @@\n'
            '+AGE_MEDIAN = 66.0000\n'
            '+WT_MEDIAN = 78.0000\n'
            '@@ -6,0 +9,3 @@\n'
            '+VWT = EXP(THETA(4)*(WT - WT_MEDIAN))\n'
            '+VAGE = EXP(THETA(5)*(AGE - AGE_MEDIAN))\n'
            '+V = V*VAGE*VWT\n',
            False,
        ),
        (
            ('nonmem', 'models', 'mox2.mod'),
            [
                ('V', 'WT', 'exp', '*'),
                ('V', 'AGE', 'exp', '*'),
                ('V', 'CLCR', 'exp', '*'),
            ],
            '@@ -1,0 +2,3 @@\n'
            '+CLCR_MEDIAN = 65.0000\n'
            '+AGE_MEDIAN = 66.0000\n'
            '+WT_MEDIAN = 78.0000\n'
            '@@ -6,0 +10,4 @@\n'
            '+VWT = EXP(THETA(4)*(WT - WT_MEDIAN))\n'
            '+VAGE = EXP(THETA(5)*(AGE - AGE_MEDIAN))\n'
            '+VCLCR = EXP(THETA(6)*(CLCR - CLCR_MEDIAN))\n'
            '+V = V*VAGE*VCLCR*VWT\n',
            False,
        ),
        (
            ('nonmem', 'models', 'mox2.mod'),
            [
                ('CL', 'CLCR', 'exp', '*'),
                ('V', 'CLCR', 'exp', '*'),
            ],
            '@@ -1,0 +2 @@\n'
            '+CLCR_MEDIAN = 65.0000\n'
            '@@ -2,0 +4,2 @@\n'
            '+CLCLCR = EXP(THETA(4)*(CLCR - CLCR_MEDIAN))\n'
            '+CL = CL*CLCLCR\n'
            '@@ -6,0 +10,2 @@\n'
            '+VCLCR = EXP(THETA(5)*(CLCR - CLCR_MEDIAN))\n'
            '+V = V*VCLCR\n',
            False,
        ),
        (
            ('nonmem', 'models', 'mox2.mod'),
            [
                ('V', 'CLCR', 'exp', '*'),
                ('CL', 'CLCR', 'exp', '*'),
            ],
            '@@ -1,0 +2 @@\n'
            '+CLCR_MEDIAN = 65.0000\n'
            '@@ -2,0 +4,2 @@\n'
            '+CLCLCR = EXP(THETA(5)*(CLCR - CLCR_MEDIAN))\n'
            '+CL = CL*CLCLCR\n'
            '@@ -6,0 +10,2 @@\n'
            '+VCLCR = EXP(THETA(4)*(CLCR - CLCR_MEDIAN))\n'
            '+V = V*VCLCR\n',
            False,
        ),
        (
            ('nonmem', 'models', 'mox2.mod'),
            [('CL', 'WT', '((cov/std) - median) * theta', '*')],
            '@@ -1,0 +2,2 @@\n'
            '+WT_MEDIAN = 78.0000\n'
            '+WT_STD = 15.6125\n'
            '@@ -2,0 +5,2 @@\n'
            '+CLWT = THETA(4)*(WT/WT_STD - WT_MEDIAN)\n'
            '+CL = CL*CLWT\n',
            False,
        ),
        (
            ('nonmem', 'models', 'mox2.mod'),
            [('MAT', 'SEX', 'cat', '*')],
            '@@ -4,0 +5,6 @@\n'
            '+IF (SEX.EQ.1) THEN\n'
            '+    MATSEX = 1\n'
            '+ELSE IF (SEX.EQ.2) THEN\n'
            '+    MATSEX = THETA(4) + 1\n'
            '+END IF\n'
            '+MAT = MAT*MATSEX\n',
            False,
        ),
        (
            ('nonmem', 'models', 'mox2.mod'),
            [('V', 'WT', 'lin', '+')],
            '@@ -1,0 +2 @@\n'
            '+WT_MEDIAN = 78.0000\n'
            '@@ -6,0 +8,2 @@\n'
            '+VWT = THETA(4)*(WT - WT_MEDIAN) + 1\n'
            '+V = V + VWT\n',
            False,
        ),
        (
            ('nonmem', 'models', 'mox2.mod'),
            [('V', 'WT', 'pow', '*')],
            '@@ -1,0 +2 @@\n'
            '+WT_MEDIAN = 78.0000\n'
            '@@ -6,0 +8,2 @@\n'
            '+VWT = (WT/WT_MEDIAN)**THETA(4)\n'
            '+V = V*VWT\n',
            False,
        ),
        (
            ('nonmem', 'models', 'mox2.mod'),
            [('V', 'WT', 'piece_lin', '+')],
            '@@ -1,0 +2 @@\n'
            '+WT_MEDIAN = 78.0000\n'
            '@@ -6,0 +8,6 @@\n'
            '+IF (WT.LE.WT_MEDIAN) THEN\n'
            '+    VWT = THETA(4)*(WT - WT_MEDIAN) + 1\n'
            '+ELSE\n'
            '+    VWT = THETA(5)*(WT - WT_MEDIAN) + 1\n'
            '+END IF\n'
            '+V = V + VWT\n',
            False,
        ),
    ],
    ids=repr,
)
def test_add_covariate_effect(
    load_model_for_test, testdata, model_path, effects, expected, allow_nested
):
    model = load_model_for_test(testdata.joinpath(*model_path))
    error_record_before = ''.join(map(str, model.internals.control_stream.get_records('ERROR')))

    if not allow_nested:
        for effect in effects:
            assert not has_covariate_effect(model, effect[0], effect[1])

    for effect in effects:
        model = add_covariate_effect(model, *effect, allow_nested=allow_nested)

    for effect in effects:
        assert has_covariate_effect(model, effect[0], effect[1])

    error_record_after = ''.join(map(str, model.internals.control_stream.get_records('ERROR')))

    original_model = load_model_for_test(testdata.joinpath(*model_path))
    assert (
        diff(
            original_model.internals.control_stream.get_pred_pk_record(),
            model.internals.control_stream.get_pred_pk_record(),
        )
        == expected
    )
    assert error_record_after == error_record_before

    for effect in effects:
        assert f'POP_{effect[0]}{effect[1]}' in model.model_code

    if not allow_nested:
        for effect in effects:
            model = remove_covariate_effect(model, effect[0], effect[1])

        for effect in effects:
            assert not has_covariate_effect(model, effect[0], effect[1])

        assert (
            diff(
                original_model.internals.control_stream.get_pred_pk_record(),
                model.internals.control_stream.get_pred_pk_record(),
            )
            == ''
        )

        for effect in effects:
            assert f'POP_{effect[0]}{effect[1]}' not in model.model_code


@pytest.mark.parametrize(
    'cov_eff,symbol,expression',
    [
        (
            CovariateEffect.exponential(),
            S('CLWGT'),
            sympy.exp(S('COVEFF1') * (S('WGT') - S('WGT_MEDIAN'))),
        ),
        (CovariateEffect.power(), S('CLWGT'), (S('WGT') / S('WGT_MEDIAN')) ** S('COVEFF1')),
        (CovariateEffect.linear(), S('CLWGT'), 1 + S('COVEFF1') * (S('WGT') - S('WGT_MEDIAN'))),
    ],
)
def test_apply(cov_eff, symbol, expression):
    cov_eff.apply(
        parameter='CL',
        covariate='WGT',
        thetas={'theta': 'COVEFF1'},
        statistics={'mean': 1, 'median': 1, 'std': 1},
    )

    assert cov_eff.template.symbol == symbol
    assert cov_eff.template.expression == expression


@pytest.mark.parametrize(
    'cov_eff, init, lower, upper', [('exp', 0.001, -0.8696, 0.8696), ('pow', 0.001, -100, 100000)]
)
def test_choose_param_inits(pheno_path, load_model_for_test, cov_eff, init, lower, upper):
    model = load_model_for_test(pheno_path)

    inits = _choose_param_inits(cov_eff, model, 'WGT')

    assert inits['init'] == init
    assert inits['lower'] == lower
    assert inits['upper'] == upper


@pytest.mark.parametrize(
    ('model_path', 'effect', 'has'),
    [
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('CL', 'PREP'),
            True,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('V1', 'PREP'),
            True,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('V2', 'PREP'),
            False,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('Q', 'PREP'),
            False,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('CL', 'AGE'),
            True,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('V1', 'AGE'),
            False,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('V2', 'AGE'),
            False,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('Q', 'AGE'),
            False,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('CL', 'OCC'),
            True,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('V1', 'OCC'),
            True,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('V2', 'OCC'),
            False,
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            ('Q', 'OCC'),
            False,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            ('CL', 'WGT'),
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            ('V', 'WGT'),
            True,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            ('CL', 'APGR'),
            False,
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            ('V', 'APGR'),
            True,
        ),
        (
            ('nonmem', 'pheno.mod'),
            ('CL', 'WGT'),
            False,
        ),
        (
            ('nonmem', 'pheno.mod'),
            ('V', 'WGT'),
            False,
        ),
        (
            ('nonmem', 'pheno.mod'),
            ('CL', 'APGR'),
            False,
        ),
        (
            ('nonmem', 'pheno.mod'),
            ('V', 'APGR'),
            False,
        ),
    ],
    ids=repr,
)
def test_has_covariate_effect(load_model_for_test, testdata, model_path, effect, has):
    model = load_model_for_test(testdata.joinpath(*model_path))

    assert has_covariate_effect(model, *effect) is has


@pytest.mark.parametrize(
    ('model_path', 'effects', 'expected'),
    [
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'WGT')],
            '@@ -4 +4 @@\n' '-TVCL=THETA(1)*WGT\n' '+TVCL = THETA(1)\n',
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('V', 'WGT')],
            '@@ -5 +5 @@\n' '-TVV=THETA(2)*WGT\n' '+TVV = THETA(2)\n',
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('V', 'APGR')],
            '@@ -6 +5,0 @@\n' '-IF(APGR.LT.5) TVV=TVV*(1+THETA(3))\n',
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('CL', 'WGT'), ('V', 'WGT')],
            '@@ -4,2 +4,2 @@\n'
            '-TVCL=THETA(1)*WGT\n'
            '-TVV=THETA(2)*WGT\n'
            '+TVCL = THETA(1)\n'
            '+TVV = THETA(2)\n',
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('V', 'WGT'), ('CL', 'WGT')],
            '@@ -4,2 +4,2 @@\n'
            '-TVCL=THETA(1)*WGT\n'
            '-TVV=THETA(2)*WGT\n'
            '+TVCL = THETA(1)\n'
            '+TVV = THETA(2)\n',
        ),
        (
            ('nonmem', 'pheno_real.mod'),
            [('V', 'WGT'), ('CL', 'WGT'), ('V', 'APGR')],
            '@@ -4,3 +4,2 @@\n'
            '-TVCL=THETA(1)*WGT\n'
            '-TVV=THETA(2)*WGT\n'
            '-IF(APGR.LT.5) TVV=TVV*(1+THETA(3))\n'
            '+TVCL = THETA(1)\n'
            '+TVV = THETA(2)\n',
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            [('CL', 'WT')],
            '@@ -29,10 +29,6 @@\n'
            '-IF (PREP2.EQ.1) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(13)*(AGE-40)) ;80 = Median of body weight\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8)) \n'
            '-ENDIF\n'
            '-IF (PREP2.EQ.2) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(12))*(1+THETA(13)*(AGE-40))\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8))*(1+THETA(12))\n'
            '-ENDIF\n'
            '-TVQ  = THETA(5)*(WT/80)**THETA(9)\n'
            '-TVV2 = THETA(6)*(WT/80)**THETA(10)\n'
            '+IF (PREP2.EQ.1) TVCL = THETA(1)*(THETA(12)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.1) TVV1 = THETA(2)*(WT/80)**THETA(7)\n'
            '+IF (PREP2.EQ.2) TVCL = THETA(1)*(THETA(11) + 1)*(THETA(12)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.2) TVV1 = THETA(2)*(WT/80)**THETA(7)*(THETA(11) + 1)\n'
            '+TVQ = THETA(5)*(WT/80)**THETA(8)\n'
            '+TVV2 = THETA(6)*(WT/80)**THETA(9)\n'
            '@@ -48 +44 @@\n'
            '-TVBA = THETA(11)\n'
            '+TVBA = THETA(10)\n',
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            [('CL', 'WT'), ('CL', 'AGE')],
            '@@ -29,10 +29,6 @@\n'
            '-IF (PREP2.EQ.1) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(13)*(AGE-40)) ;80 = Median of body weight\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8)) \n'
            '-ENDIF\n'
            '-IF (PREP2.EQ.2) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(12))*(1+THETA(13)*(AGE-40))\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8))*(1+THETA(12))\n'
            '-ENDIF\n'
            '-TVQ  = THETA(5)*(WT/80)**THETA(9)\n'
            '-TVV2 = THETA(6)*(WT/80)**THETA(10)\n'
            '+IF (PREP2.EQ.1) TVCL = THETA(1)\n'
            '+IF (PREP2.EQ.1) TVV1 = THETA(2)*(WT/80)**THETA(7)\n'
            '+IF (PREP2.EQ.2) TVCL = THETA(1)*(THETA(11) + 1)\n'
            '+IF (PREP2.EQ.2) TVV1 = THETA(2)*(WT/80)**THETA(7)*(THETA(11) + 1)\n'
            '+TVQ = THETA(5)*(WT/80)**THETA(8)\n'
            '+TVV2 = THETA(6)*(WT/80)**THETA(9)\n'
            '@@ -48 +44 @@\n'
            '-TVBA = THETA(11)\n'
            '+TVBA = THETA(10)\n',
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            [('V1', 'WT')],
            '@@ -29,10 +29,6 @@\n'
            '-IF (PREP2.EQ.1) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(13)*(AGE-40)) ;80 = Median of body weight\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8)) \n'
            '-ENDIF\n'
            '-IF (PREP2.EQ.2) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(12))*(1+THETA(13)*(AGE-40))\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8))*(1+THETA(12))\n'
            '-ENDIF\n'
            '-TVQ  = THETA(5)*(WT/80)**THETA(9)\n'
            '-TVV2 = THETA(6)*(WT/80)**THETA(10)\n'
            '+IF (PREP2.EQ.1) TVCL = THETA(1)*(WT/80)**THETA(7)*(THETA(12)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.1) TVV1 = THETA(2)\n'
            '+IF (PREP2.EQ.2) TVCL = THETA(1)*(WT/80)**THETA(7)*(THETA(11) + 1)*(THETA(12)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.2) TVV1 = THETA(2)*(THETA(11) + 1)\n'
            '+TVQ = THETA(5)*(WT/80)**THETA(8)\n'
            '+TVV2 = THETA(6)*(WT/80)**THETA(9)\n'
            '@@ -48 +44 @@\n'
            '-TVBA = THETA(11)\n'
            '+TVBA = THETA(10)\n',
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            [('V2', 'WT')],
            '@@ -29,8 +29,4 @@\n'
            '-IF (PREP2.EQ.1) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(13)*(AGE-40)) ;80 = Median of body weight\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8)) \n'
            '-ENDIF\n'
            '-IF (PREP2.EQ.2) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(12))*(1+THETA(13)*(AGE-40))\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8))*(1+THETA(12))\n'
            '-ENDIF\n'
            '+IF (PREP2.EQ.1) TVCL = THETA(1)*(WT/80)**THETA(7)*(THETA(12)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.1) TVV1 = THETA(2)*(WT/80)**THETA(8)\n'
            '+IF (PREP2.EQ.2) TVCL = THETA(1)*(WT/80)**THETA(7)*(THETA(11) + 1)*(THETA(12)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.2) TVV1 = THETA(2)*(WT/80)**THETA(8)*(THETA(11) + 1)\n'
            '@@ -38 +34 @@\n'
            '-TVV2 = THETA(6)*(WT/80)**THETA(10)\n'
            '+TVV2 = THETA(6)\n'
            '@@ -48 +44 @@\n'
            '-TVBA = THETA(11)\n'
            '+TVBA = THETA(10)\n',
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            [('Q', 'WT')],
            '@@ -29,10 +29,6 @@\n'
            '-IF (PREP2.EQ.1) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(13)*(AGE-40)) ;80 = Median of body weight\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8)) \n'
            '-ENDIF\n'
            '-IF (PREP2.EQ.2) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(12))*(1+THETA(13)*(AGE-40))\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8))*(1+THETA(12))\n'
            '-ENDIF\n'
            '-TVQ  = THETA(5)*(WT/80)**THETA(9)\n'
            '-TVV2 = THETA(6)*(WT/80)**THETA(10)\n'
            '+IF (PREP2.EQ.1) TVCL = THETA(1)*(WT/80)**THETA(7)*(THETA(12)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.1) TVV1 = THETA(2)*(WT/80)**THETA(8)\n'
            '+IF (PREP2.EQ.2) TVCL = THETA(1)*(WT/80)**THETA(7)*(THETA(11) + 1)*(THETA(12)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.2) TVV1 = THETA(2)*(WT/80)**THETA(8)*(THETA(11) + 1)\n'
            '+TVQ = THETA(5)\n'
            '+TVV2 = THETA(6)*(WT/80)**THETA(9)\n'
            '@@ -48 +44 @@\n'
            '-TVBA = THETA(11)\n'
            '+TVBA = THETA(10)\n',
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            [('K21', 'WT')],
            '@@ -29,10 +29,6 @@\n'
            '-IF (PREP2.EQ.1) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(13)*(AGE-40)) ;80 = Median of body weight\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8)) \n'
            '-ENDIF\n'
            '-IF (PREP2.EQ.2) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(12))*(1+THETA(13)*(AGE-40))\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8))*(1+THETA(12))\n'
            '-ENDIF\n'
            '-TVQ  = THETA(5)*(WT/80)**THETA(9)\n'
            '-TVV2 = THETA(6)*(WT/80)**THETA(10)\n'
            '+IF (PREP2.EQ.1) TVCL = THETA(1)*(WT/80)**THETA(7)*(THETA(11)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.1) TVV1 = THETA(2)*(WT/80)**THETA(8)\n'
            '+IF (PREP2.EQ.2) TVCL = THETA(1)*(WT/80)**THETA(7)*(THETA(10) + 1)*(THETA(11)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.2) TVV1 = THETA(2)*(WT/80)**THETA(8)*(THETA(10) + 1)\n'
            '+TVQ = THETA(5)\n'
            '+TVV2 = THETA(6)\n'
            '@@ -48 +44 @@\n'
            '-TVBA = THETA(11)\n'
            '+TVBA = THETA(9)\n',
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            [('CL', 'WT'), ('V1', 'WT')],
            '@@ -29,10 +29,6 @@\n'
            '-IF (PREP2.EQ.1) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(13)*(AGE-40)) ;80 = Median of body weight\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8)) \n'
            '-ENDIF\n'
            '-IF (PREP2.EQ.2) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(12))*(1+THETA(13)*(AGE-40))\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8))*(1+THETA(12))\n'
            '-ENDIF\n'
            '-TVQ  = THETA(5)*(WT/80)**THETA(9)\n'
            '-TVV2 = THETA(6)*(WT/80)**THETA(10)\n'
            '+IF (PREP2.EQ.1) TVCL = THETA(1)*(THETA(11)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.1) TVV1 = THETA(2)\n'
            '+IF (PREP2.EQ.2) TVCL = THETA(1)*(THETA(10) + 1)*(THETA(11)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.2) TVV1 = THETA(2)*(THETA(10) + 1)\n'
            '+TVQ = THETA(5)*(WT/80)**THETA(7)\n'
            '+TVV2 = THETA(6)*(WT/80)**THETA(8)\n'
            '@@ -48 +44 @@\n'
            '-TVBA = THETA(11)\n'
            '+TVBA = THETA(9)\n',
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            [('V1', 'WT'), ('CL', 'AGE')],
            '@@ -29,10 +29,6 @@\n'
            '-IF (PREP2.EQ.1) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(13)*(AGE-40)) ;80 = Median of body weight\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8)) \n'
            '-ENDIF\n'
            '-IF (PREP2.EQ.2) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(12))*(1+THETA(13)*(AGE-40))\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8))*(1+THETA(12))\n'
            '-ENDIF\n'
            '-TVQ  = THETA(5)*(WT/80)**THETA(9)\n'
            '-TVV2 = THETA(6)*(WT/80)**THETA(10)\n'
            '+IF (PREP2.EQ.1) TVCL = THETA(1)*(WT/80)**THETA(7)\n'
            '+IF (PREP2.EQ.1) TVV1 = THETA(2)\n'
            '+IF (PREP2.EQ.2) TVCL = THETA(1)*(WT/80)**THETA(7)*(THETA(11) + 1)\n'
            '+IF (PREP2.EQ.2) TVV1 = THETA(2)*(THETA(11) + 1)\n'
            '+TVQ = THETA(5)*(WT/80)**THETA(8)\n'
            '+TVV2 = THETA(6)*(WT/80)**THETA(9)\n'
            '@@ -48 +44 @@\n'
            '-TVBA = THETA(11)\n'
            '+TVBA = THETA(10)\n',
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            [('CL', 'WT'), ('V1', 'WT'), ('V2', 'WT')],
            '@@ -29,10 +29,6 @@\n'
            '-IF (PREP2.EQ.1) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(13)*(AGE-40)) ;80 = Median of body weight\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8)) \n'
            '-ENDIF\n'
            '-IF (PREP2.EQ.2) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(12))*(1+THETA(13)*(AGE-40))\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8))*(1+THETA(12))\n'
            '-ENDIF\n'
            '-TVQ  = THETA(5)*(WT/80)**THETA(9)\n'
            '-TVV2 = THETA(6)*(WT/80)**THETA(10)\n'
            '+IF (PREP2.EQ.1) TVCL = THETA(1)*(THETA(10)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.1) TVV1 = THETA(2)\n'
            '+IF (PREP2.EQ.2) TVCL = THETA(1)*(THETA(9) + 1)*(THETA(10)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.2) TVV1 = THETA(2)*(THETA(9) + 1)\n'
            '+TVQ = THETA(5)*(WT/80)**THETA(7)\n'
            '+TVV2 = THETA(6)\n'
            '@@ -48 +44 @@\n'
            '-TVBA = THETA(11)\n'
            '+TVBA = THETA(8)\n',
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            [('CL', 'WT'), ('V1', 'WT'), ('Q', 'WT'), ('V2', 'WT')],
            '@@ -29,10 +29,6 @@\n'
            '-IF (PREP2.EQ.1) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(13)*(AGE-40)) ;80 = Median of body weight\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8)) \n'
            '-ENDIF\n'
            '-IF (PREP2.EQ.2) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(12))*(1+THETA(13)*(AGE-40))\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8))*(1+THETA(12))\n'
            '-ENDIF\n'
            '-TVQ  = THETA(5)*(WT/80)**THETA(9)\n'
            '-TVV2 = THETA(6)*(WT/80)**THETA(10)\n'
            '+IF (PREP2.EQ.1) TVCL = THETA(1)*(THETA(9)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.1) TVV1 = THETA(2)\n'
            '+IF (PREP2.EQ.2) TVCL = THETA(1)*(THETA(8) + 1)*(THETA(9)*(AGE - 40) + 1)\n'
            '+IF (PREP2.EQ.2) TVV1 = THETA(2)*(THETA(8) + 1)\n'
            '+TVQ = THETA(5)\n'
            '+TVV2 = THETA(6)\n'
            '@@ -48 +44 @@\n'
            '-TVBA = THETA(11)\n'
            '+TVBA = THETA(7)\n',
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            [('CL', 'WT'), ('V1', 'WT'), ('V2', 'WT'), ('CL', 'AGE')],
            '@@ -29,10 +29,6 @@\n'
            '-IF (PREP2.EQ.1) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(13)*(AGE-40)) ;80 = Median of body weight\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8)) \n'
            '-ENDIF\n'
            '-IF (PREP2.EQ.2) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(12))*(1+THETA(13)*(AGE-40))\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8))*(1+THETA(12))\n'
            '-ENDIF\n'
            '-TVQ  = THETA(5)*(WT/80)**THETA(9)\n'
            '-TVV2 = THETA(6)*(WT/80)**THETA(10)\n'
            '+IF (PREP2.EQ.1) TVCL = THETA(1)\n'
            '+IF (PREP2.EQ.1) TVV1 = THETA(2)\n'
            '+IF (PREP2.EQ.2) TVCL = THETA(1)*(THETA(9) + 1)\n'
            '+IF (PREP2.EQ.2) TVV1 = THETA(2)*(THETA(9) + 1)\n'
            '+TVQ = THETA(5)*(WT/80)**THETA(7)\n'
            '+TVV2 = THETA(6)\n'
            '@@ -48 +44 @@\n'
            '-TVBA = THETA(11)\n'
            '+TVBA = THETA(8)\n',
        ),
        (
            ('nonmem', 'models', 'fviii6.mod'),
            [('CL', 'WT'), ('V1', 'WT'), ('V2', 'WT'), ('CL', 'AGE'), ('Q', 'WT')],
            '@@ -29,10 +29,6 @@\n'
            '-IF (PREP2.EQ.1) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(13)*(AGE-40)) ;80 = Median of body weight\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8)) \n'
            '-ENDIF\n'
            '-IF (PREP2.EQ.2) THEN\n'
            '-TVCL = THETA(1)*((WT/80)**THETA(7))*(1+THETA(12))*(1+THETA(13)*(AGE-40))\n'
            '-TVV1 = THETA(2)*((WT/80)**THETA(8))*(1+THETA(12))\n'
            '-ENDIF\n'
            '-TVQ  = THETA(5)*(WT/80)**THETA(9)\n'
            '-TVV2 = THETA(6)*(WT/80)**THETA(10)\n'
            '+IF (PREP2.EQ.1) TVCL = THETA(1)\n'
            '+IF (PREP2.EQ.1) TVV1 = THETA(2)\n'
            '+IF (PREP2.EQ.2) TVCL = THETA(1)*(THETA(8) + 1)\n'
            '+IF (PREP2.EQ.2) TVV1 = THETA(2)*(THETA(8) + 1)\n'
            '+TVQ = THETA(5)\n'
            '+TVV2 = THETA(6)\n'
            '@@ -48 +44 @@\n'
            '-TVBA = THETA(11)\n'
            '+TVBA = THETA(7)\n',
        ),
    ],
    ids=repr,
)
def test_remove_covariate_effect(load_model_for_test, testdata, model_path, effects, expected):
    model = load_model_for_test(testdata.joinpath(*model_path))
    error_record_before = ''.join(map(str, model.internals.control_stream.get_records('ERROR')))

    for effect in effects:
        assert has_covariate_effect(model, effect[0], effect[1])

    for effect in effects:
        model = remove_covariate_effect(model, *effect)

    for effect in effects:
        assert not has_covariate_effect(model, effect[0], effect[1])

    error_record_after = ''.join(map(str, model.internals.control_stream.get_records('ERROR')))

    original_model = load_model_for_test(testdata.joinpath(*model_path))
    assert (
        diff(
            original_model.internals.control_stream.get_pred_pk_record(),
            model.internals.control_stream.get_pred_pk_record(),
        )
        == expected
    )
    assert error_record_after == error_record_before



from pharmpy.internals.module.lazy import LazyImport

box = LazyImport('box', globals(), 'rich.box')
console = LazyImport('console', globals(), 'rich.console')
table = LazyImport('table', globals(), 'rich.table')



import pandas as pd
import pytest
import sympy

from pharmpy.tools import read_modelfit_results
from pharmpy.tools.structsearch.drugmetabolite import create_drug_metabolite_models
from pharmpy.tools.structsearch.pkpd import create_pkpd_models
from pharmpy.tools.structsearch.tmdd import (
    create_cr_models,
    create_crib_models,
    create_full_models,
    create_ib_models,
    create_mmapp_model,
    create_qss_models,
    create_remaining_models,
    create_wagner_model,
)
from pharmpy.tools.structsearch.tool import create_workflow, validate_input
from pharmpy.workflows import Workflow

ests = pd.Series(
    {
        'POP_R_0': 1.0,
        'IIV_R_0': 6.0,
        'POP_KDC': 2.0,
        'POP_KINT': 3.0,
        'POP_KDEG': 4.0,
        'POP_VC': 5.0,
    }
)


def test_create_qss_models(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    models = create_qss_models(model, ests, None)
    assert len(models) == 8


def test_create_qss_models_multiple_dvs(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    models = create_qss_models(model, ests, {'target': 3, 'complex': 2})
    assert len(models) == 8
    assert models[0].dependent_variables == {
        sympy.Symbol('Y'): 1,
        sympy.Symbol('Y_TARGET'): 3,
        sympy.Symbol('Y_COMPLEX'): 2,
    }


def test_create_wagner_model(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    models = create_wagner_model(model, ests, None)
    assert len(models) == 1


def test_create_mmapp_model(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    models = create_mmapp_model(model, ests, None)
    assert len(models) == 1


def test_create_crib_models(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    models = create_crib_models(model, ests, None)
    assert len(models) == 2


def test_create_full_models(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    models = create_full_models(model, ests, None)
    assert len(models) == 4


def test_create_ib_models(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    models = create_ib_models(model, ests, None)
    assert len(models) == 2


def test_create_cr_models(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    models = create_cr_models(model, ests, None)
    assert len(models) == 2


def test_create_remaining_models(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    models = create_remaining_models(model, ests, 2, None)
    assert len(models) == 12


def test_create_remaining_models_multiple_dvs(load_example_model_for_test):
    model = load_example_model_for_test("pheno")
    models = create_remaining_models(model, ests, 2, {'target': 2, 'complex': 3})
    assert len(models) == 12


def test_pkpd(load_model_for_test, testdata):
    search_space = "DIRECTEFFECT(*); EFFECTCOMP(*); INDIRECTEFFECT(*,*)"
    res = read_modelfit_results(testdata / "nonmem" / "pheno.mod")
    ests = res.parameter_estimates
    model = load_model_for_test(testdata / "nonmem" / "pheno_pd.mod")
    pkpd_models = create_pkpd_models(
        model, search_space, b_init=5.75, ests=ests, emax_init=2.0, ec50_init=1.0, met_init=0.5
    )

    assert len(pkpd_models) == 12
    assert pkpd_models[0].name == "structsearch_run1"
    assert pkpd_models[1].name == "structsearch_run2"
    assert pkpd_models[1].parameters['POP_B'].init == 5.75
    assert pkpd_models[1].parameters['POP_E_MAX'].init == 2.0
    assert pkpd_models[1].parameters['POP_E_MAX'].fix is False
    assert pkpd_models[1].parameters['POP_EC_50'].init == 1.0
    assert pkpd_models[1].parameters['POP_EC_50'].fix is False
    assert pkpd_models[3].parameters['POP_MET'].init == 0.5
    assert pkpd_models[6].parameters['POP_MET'].init == 0.5

    models3 = create_pkpd_models(model, search_space)
    assert models3[1].parameters['POP_E_MAX'].init == 0.1
    assert models3[1].parameters['POP_E_MAX'].fix is False
    assert models3[1].parameters['POP_EC_50'].init == 0.1
    assert models3[1].parameters['POP_EC_50'].fix is False


def test_drug_metabolite(load_model_for_test, testdata):
    model = load_model_for_test(testdata / "nonmem" / "pheno_pd.mod")
    res = read_modelfit_results(testdata / "nonmem" / "pheno.mod")
    search_space = "METABOLITE([PSC, BASIC]);PERIPHERALS([0,1], MET)"
    wb, candidate_tasks, base_model_description = create_drug_metabolite_models(
        model, res, search_space
    )
    assert base_model_description == "METABOLITE_BASIC;PERIPHERALS(0)"
    assert len(candidate_tasks) == 4

    wb, candidate_tasks, base_model_description = create_drug_metabolite_models(
        model, res, "METABOLITE([PSC, BASIC])"
    )
    assert base_model_description == "METABOLITE_BASIC"
    assert len(candidate_tasks) == 2

    wb, candidate_tasks, base_model_description = create_drug_metabolite_models(
        model, res, "METABOLITE(BASIC);PERIPHERALS([0,1], MET)"
    )
    assert base_model_description == "METABOLITE_BASIC;PERIPHERALS(0)"
    assert len(candidate_tasks) == 2


def test_create_workflow():
    assert isinstance(create_workflow('pkpd'), Workflow)


def test_create_workflow_pkpd(load_model_for_test, testdata):
    model = load_model_for_test(testdata / "nonmem" / "pheno_pd.mod")
    assert isinstance(create_workflow('pkpd', model=model), Workflow)


def test_create_workflow_tmdd(load_model_for_test, testdata):
    model = load_model_for_test(testdata / "nonmem" / "pheno_pd.mod")
    assert isinstance(create_workflow('pkpd', model=model), Workflow)


def test_create_workflow_drug_metabolite(load_model_for_test, testdata):
    model = load_model_for_test(testdata / "nonmem" / "pheno_pd.mod")
    assert isinstance(create_workflow('drug_metabolite', model=model), Workflow)


@pytest.mark.parametrize(
    ('arguments', 'exception', 'match'),
    [
        (
            dict(type='tmdd', emax_init=0.2),
            ValueError,
            'Invalid arguments "b_init", "emax_init", "ec50_init" and "met_init" for TMDD models.',
        ),
        (
            dict(type='a'),
            ValueError,
            'Invalid `type`',
        ),
        (
            dict(type='tmdd', search_space='ABSORPTION'),
            ValueError,
            'Invalid argument "search_space" for TMDD models.',
        ),
        (
            dict(type='pkpd', dv_types={'drug': 1}),
            ValueError,
            'Invalid argument "dv_types" for PKPD models.',
        ),
        (
            dict(type='drug_metabolite', dv_types={'drug': 1}),
            ValueError,
            'Invalid argument "dv_types" for drug metabolite models.',
        ),
        (
            dict(type='drug_metabolite', met_init=1),
            ValueError,
            'Invalid arguments "b_init", "emax_init", "ec50_init" and "met_init" for drug metabolite models.',
        ),
    ],
)
def test_validation(tmp_path, load_model_for_test, testdata, arguments, exception, match):
    model = load_model_for_test(testdata / "nonmem" / "pheno.mod")
    res = read_modelfit_results(testdata / "nonmem" / "pheno.mod")

    kwargs = {**arguments}
    with pytest.raises(exception, match=match):
        validate_input(**kwargs)

    with pytest.raises(exception, match='Invalid argument "extra_model" for PKPD models.'):
        validate_input(type='pkpd', extra_model=model)
    with pytest.raises(exception, match='Invalid argument "extra_model_results" for PKPD models.'):
        validate_input(type='pkpd', extra_model_results=res)
    with pytest.raises(
        exception, match='Invalid argument "extra_model" for drug metabolite models.'
    ):
        validate_input(type='drug_metabolite', extra_model=model)
    with pytest.raises(
        exception, match='Invalid argument "extra_model_results" for drug metabolite models.'
    ):
        validate_input(type='drug_metabolite', extra_model_results=res)

    validate_input(type='tmdd', dv_types={'drug_tot': 1, 'target_tot': 2, 'complex': 3})
    validate_input(type='tmdd', dv_types={'drug': 1, 'target_tot': 2, 'complex': 3})
    validate_input(type='tmdd', dv_types={'drug': 1, 'target': 2, 'complex': 3})

    with pytest.raises(exception, match='Values must be unique.'):
        validate_input(type='tmdd', dv_types={'drug': 1, 'target': 1, 'complex': 2})
    with pytest.raises(exception, match='Only drug can have DVID = 1. Please choose another DVID.'):
        validate_input(type='tmdd', dv_types={'target': 1, 'complex': 2})



from abc import ABC, abstractmethod
from pathlib import Path


class ToolDatabase(ABC):
    """Database of results, metadata and run files for one tool run

    Parameters
    ----------
    toolname : str
        Name of the tool to use the database
    """

    def __init__(self, toolname):
        self.toolname = toolname

    @property
    def path(self):
        """ModelDatabase to store results of models run by tool"""
        return self._path

    @path.setter
    def path(self, value: Path):
        self._path = value

    @property
    def model_database(self):
        """ModelDatabase to store results of models run by tool"""
        return self._model_database

    @model_database.setter
    def model_database(self, value):
        self._model_database = value

    @abstractmethod
    def store_local_file(self, source_path):
        """Store a local file in the database

        Parameters
        ----------
        source_path : str or Path
            Path to local file
        """
        pass

    @abstractmethod
    def store_results(self, res):
        """Store tool results

        Parameters
        ----------
        res : Results
            Tool results object
        """
        pass

    @abstractmethod
    def store_metadata(self, metadata):
        """Store tool metadata

        Parameters
        ----------
        metadata : dict
            Tool metadata dictionary
        """
        pass

    @abstractmethod
    def read_metadata(self):
        """Read tool metadata"""
        pass



import pytest
from sympy import Symbol as S

from pharmpy.modeling import add_iov, fix_parameters, remove_iov
from pharmpy.tools.iovsearch.tool import (
    _get_iiv_etas_with_corresponding_iov,
    _get_nonfixed_iivs,
    create_workflow,
    validate_input,
)
from pharmpy.workflows import Workflow


def test_ignore_fixed_iiv(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    assert len(model.random_variables.iiv.names) == 2
    model = fix_parameters(model, 'IVCL')
    nonfixed_iivs = _get_nonfixed_iivs(model)
    assert len(nonfixed_iivs.names) == 1


def test_iovsearch_github_issues_976(load_model_for_test, testdata):
    m = load_model_for_test(testdata / 'nonmem' / 'pheno_multivariate_piecewise.mod')
    assert not m.random_variables.iov
    assert set(_get_iiv_etas_with_corresponding_iov(m)) == set()

    m = add_iov(m, 'FA1', distribution='same-as-iiv')
    assert set(_get_iiv_etas_with_corresponding_iov(m)) == set(
        map(lambda rv: S(rv), m.random_variables.iiv.names)
    )

    m = remove_iov(m, 'ETA_IOV_1_1')
    assert set(_get_iiv_etas_with_corresponding_iov(m)) == {S('ETA_2')}

    m = remove_iov(m, 'ETA_IOV_2_1')
    assert not m.random_variables.iov
    assert set(_get_iiv_etas_with_corresponding_iov(m)) == set()


def test_create_workflow():
    assert isinstance(create_workflow(), Workflow)


def test_create_workflow_with_model(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    assert isinstance(create_workflow(model=model, column='APGR'), Workflow)


def test_validate_input():
    validate_input()


def test_validate_input_with_model(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    validate_input(model=model, column='APGR')


def test_validate_input_with_model_and_list_of_parameters(load_model_for_test, testdata):
    model = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    validate_input(model=model, column='APGR', list_of_parameters=['CL', 'V'])


@pytest.mark.parametrize(
    (
        'model_path',
        'arguments',
        'exception',
        'match',
    ),
    [
        (None, dict(column=1), TypeError, 'Invalid `column`'),
        (None, dict(list_of_parameters='CL'), TypeError, 'Invalid `list_of_parameters`'),
        (None, dict(rank_type=1), ValueError, 'Invalid `rank_type`'),
        (None, dict(rank_type='bi'), ValueError, 'Invalid `rank_type`'),
        (None, dict(cutoff='1'), TypeError, 'Invalid `cutoff`'),
        (None, dict(distribution=['same-as-iiv']), ValueError, 'Invalid `distribution`'),
        (None, dict(distribution='same'), ValueError, 'Invalid `distribution`'),
        (
            ('nonmem', 'pheno.mod'),
            dict(column='OCC'),
            ValueError,
            'Invalid `column`',
        ),
        (
            ('nonmem', 'pheno.mod'),
            dict(column='APGR', list_of_parameters=['K']),
            ValueError,
            'Invalid `list_of_parameters`',
        ),
        (
            None,
            dict(model=1),
            TypeError,
            'Invalid `model`',
        ),
    ],
)
def test_validate_input_raises(
    load_model_for_test,
    testdata,
    model_path,
    arguments,
    exception,
    match,
):
    model = load_model_for_test(testdata.joinpath(*model_path)) if model_path else None

    kwargs = {'model': model, **arguments}

    with pytest.raises(exception, match=match):
        validate_input(**kwargs)



from __future__ import annotations

from functools import lru_cache
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Mapping

if TYPE_CHECKING:
    import sympy
else:
    from pharmpy.deps import sympy

from .tree import replace_root_children


def subs(expr: sympy.Expr, mapping: Mapping[Any, Any], simultaneous: bool = False) -> sympy.Expr:
    _mapping = xreplace_dict(mapping)
    if (simultaneous or _mapping_is_not_recursive(_mapping)) and all(
        map(_old_does_not_need_generic_subs, _mapping.keys())
    ):
        if sympy.exp in _mapping:
            new_base = _mapping[sympy.exp]
            _mapping[sympy.exp] = lambda x: new_base**x
            _mapping[sympy.Pow] = lambda a, x: new_base**x if a is sympy.S.Exp1 else a**x
            return _subs_atoms_simultaneously(_subs_atom_or_func(_mapping), expr)
        return _subs_atoms_simultaneously(_subs_atom(_mapping), expr)
    return expr.subs(
        _mapping, simultaneous=simultaneous
    )  # pyright: ignore [reportGeneralTypeIssues]


def xreplace_dict(dictlike) -> Dict[Any, Any]:
    return {_sympify_old(key): _sympify_new(value) for key, value in dictlike.items()}


def _sympify_old(old) -> sympy.Expr:
    # NOTE: This mimics sympy's input coercion in subs
    return (
        sympy.Symbol(old)
        if isinstance(old, str)
        else sympy.sympify(old, strict=not isinstance(old, type))
    )


@lru_cache(maxsize=256)
def _sympify_new(new) -> sympy.Expr:
    # NOTE: This mimics sympy's input coercion in subs
    return sympy.sympify(new, strict=not isinstance(new, (str, type)))


def _mapping_is_not_recursive(mapping: Dict[sympy.Expr, sympy.Expr]):
    return set(mapping.keys()).isdisjoint(
        set().union(*map(lambda e: e.free_symbols, mapping.values()))
    )


def _old_does_not_need_generic_subs(expr: sympy.Expr):
    return isinstance(expr, sympy.Symbol) or expr is sympy.exp


def _subs_atoms_simultaneously(
    subs_new_args: Callable[[sympy.Expr, List[sympy.Expr]], sympy.Expr], expr: sympy.Expr
):
    stack = [expr]
    output = [[], []]

    while stack:
        e = stack[-1]
        old_args = e.args

        new_args = output[-1]

        n = len(old_args)
        i = len(new_args)

        if i == n:
            stack.pop()
            output.pop()
            output[-1].append(subs_new_args(e, new_args))
        else:
            # NOTE: Push the next argument on the stack
            stack.append(old_args[i])  # pyright: ignore [reportGeneralTypeIssues]
            output.append([])

    return output[0][0]


def _subs_atom(mapping: Dict[sympy.Expr, sympy.Expr]):
    def _subs(expr: sympy.Expr, args: List[sympy.Expr]):
        return replace_root_children(expr, args) if args else mapping.get(expr, expr)

    return _subs


def _subs_atom_or_func(mapping: Dict[Any, Any]):
    def _subs(expr: sympy.Expr, args: List[Any]):
        if not args:
            return mapping.get(expr, expr)

        fn = expr.func
        new_fn = mapping.get(fn, fn)
        return replace_root_children(expr, args) if fn is new_fn else new_fn(*args)

    return _subs



import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.modeling import calculate_parameters_from_ucp, calculate_ucp_scale, update_inits
from pharmpy.tools import run_retries


@pytest.mark.parametrize(
    ('scale',),
    (('UCP',),),
)
def test_retries(tmp_path, model_count, scale, start_modelres):
    with chdir(tmp_path):
        fraction = 0.1
        number_of_candidates = 2
        res = run_retries(
            number_of_candidates=number_of_candidates,
            fraction=fraction,
            scale=scale,
            results=start_modelres[1],
            model=start_modelres[0],
        )

        # All candidate models + start model
        assert len(res.summary_tool) == 3
        assert len(res.summary_models) == 3
        assert len(res.models) == 3
        for model in res.models:
            if model != start_modelres[0]:
                is_within_fraction(start_modelres[0], start_modelres[1], model, scale, fraction)
        rundir = tmp_path / 'retries_dir1'
        assert rundir.is_dir()
        assert model_count(rundir) == 2  # Not the start model
        assert (rundir / 'results.json').exists()
        assert (rundir / 'results.csv').exists()
        assert (rundir / 'metadata.json').exists()


def is_within_fraction(start_model, start_model_res, candidate_model, scale, fraction):
    parameter_value = list(start_model_res.parameter_estimates.items())

    allowed_dict = {}
    if scale == "normal":
        for parameter, value in parameter_value:
            allowed_dict[parameter] = (
                value - value * fraction,
                value + value * fraction,
            )
    elif scale == "UCP":
        start_model = update_inits(start_model, start_model_res.parameter_estimates)
        ucp_scale = calculate_ucp_scale(start_model)
        lower = {}
        upper = {}
        for parameter, _ in parameter_value:
            lower[parameter] = 0.1 - (0.1 * fraction)
            upper[parameter] = 0.1 + (0.1 * fraction)
        new_lower_parameters = calculate_parameters_from_ucp(start_model, ucp_scale, lower)
        new_upper_parameters = calculate_parameters_from_ucp(start_model, ucp_scale, upper)
        for parameter, _ in parameter_value:
            allowed_dict[parameter] = (
                new_lower_parameters[parameter],
                new_upper_parameters[parameter],
            )
    for parameter in candidate_model.parameters:
        assert allowed_dict[parameter.name][0] < parameter.init < allowed_dict[parameter.name][1]



from typing import TypeVar

from .context import insert_context
from .workflow import Workflow, WorkflowBuilder

T = TypeVar('T')


def call_workflow(wf: Workflow[T], unique_name, db) -> T:
    """Dynamically call a workflow from another workflow.

    Currently only supports dask distributed

    Parameters
    ----------
    wf : Workflow
        A workflow object
    unique_name : str
        A name of the results node that is unique between parent and dynamically created workflows
    db : ToolDatabase
        ToolDatabase to pass to new workflow

    Returns
    -------
    Any
        Whatever the dynamic workflow returns
    """
    from dask.distributed import get_client, rejoin, secede

    from .optimize import optimize_task_graph_for_dask_distributed

    wb = WorkflowBuilder(wf)
    insert_context(wb, db)
    wf = Workflow(wb)

    client = get_client()
    dsk = wf.as_dask_dict()
    dsk[unique_name] = dsk.pop('results')
    dsk_optimized = optimize_task_graph_for_dask_distributed(client, dsk)
    futures = client.get(dsk_optimized, unique_name, sync=False)
    secede()
    res: T = client.gather(futures)  # pyright: ignore [reportGeneralTypeIssues]
    rejoin()
    return res



import os
import re
import subprocess

from pharmpy.modeling import read_model, write_csv, write_model
from pharmpy.tools import read_results
from pharmpy.workflows import default_tool_database


def have_scm():
    # Check if scm is available through PsN
    try:
        out = subprocess.run(["scm", "--version"], capture_output=True)
    except FileNotFoundError:
        return False

    m = re.match(r"PsN version: (\d+)\.(\d+)\.(\d+)", out.stdout.decode("utf-8"))
    if m:
        major = int(m.group(1))
        minor = int(m.group(2))
        patch = int(m.group(3))
        return (
            major > 5 or (major == 5 and minor > 2) or (major == 5 and minor == 2 and patch >= 10)
        )
    else:
        return False


def run_scm(model, relations, continuous=None, categorical=None, path=None):
    if continuous is None:
        continuous = []
    if categorical is None:
        categorical = []

    if path is not None:
        path = path / 'scm'

    db = default_tool_database(toolname='scm', path=path)
    path = db.path / "psn-wrapper"
    path.mkdir()

    config = _create_config(model, path, relations, continuous, categorical)
    config_path = path / "config.scm"
    with open(config_path, 'w') as fh:
        fh.write(config)

    write_csv(model, path=path)
    write_model(model, path=path)

    os.system(f"scm {config_path} -directory={path / 'scmdir'} -auto_tv")

    res = read_results(path / 'scmdir' / 'results.json')
    final_model_path = path / 'scmdir' / 'final_models' / 'final_forward.mod'
    if final_model_path.is_file():
        final_model = read_model(final_model_path)
        res.final_model = final_model
    return res


def _create_config(model, path, relations, continuous, categorical):
    config = f"model={path}/{model.name}{model.filename_extension}\n"
    config += "search_direction=forward\n"
    config += "p_forward=0.01\n"
    if continuous:
        config += f"continuous_covariates={','.join(continuous)}\n"
    if categorical:
        config += f"categorical_covariates={','.join(categorical)}\n"
    config += f"do_not_drop={','.join(continuous + categorical)}\n"
    config += "\n"
    config += "[test_relations]\n"
    for param, covs in relations.items():
        config += f"{param}={','.join(covs)}\n"
    config += "\n"
    config += "[valid_states]\n"
    config += "continuous=1,4\n"
    config += "categorical=1,2\n"
    return config



from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import sympy
else:
    from pharmpy.deps import sympy


def is_zero_matrix(A: sympy.Matrix) -> bool:
    for e in A:
        assert isinstance(e, sympy.Expr)
        if not e.is_zero:
            return False
    return True



from pharmpy.tools.mfl.statement.definition import Let
from pharmpy.tools.mfl.statement.feature.absorption import Absorption
from pharmpy.tools.mfl.statement.feature.covariate import Covariate
from pharmpy.tools.mfl.statement.feature.direct_effect import DirectEffect
from pharmpy.tools.mfl.statement.feature.effect_comp import EffectComp
from pharmpy.tools.mfl.statement.feature.elimination import Elimination
from pharmpy.tools.mfl.statement.feature.indirect_effect import IndirectEffect
from pharmpy.tools.mfl.statement.feature.lagtime import LagTime
from pharmpy.tools.mfl.statement.feature.metabolite import Metabolite
from pharmpy.tools.mfl.statement.feature.peripherals import Peripherals
from pharmpy.tools.mfl.statement.feature.transits import Transits

from .statement.feature.symbols import Name

MODELSEARCH_STATEMENT_TYPES = (
    Absorption,
    Elimination,
    LagTime,
    Transits,
)

COVSEARCH_STATEMENT_TYPES = (
    Let,
    Covariate,
)


STRUCTSEARCH_STATEMENT_TYPES = (DirectEffect, EffectComp, IndirectEffect, Metabolite)


def mfl_filtering(statements, tool_name):
    assert tool_name in ["structsearch", "modelsearch"]

    if tool_name == "modelsearch":
        statement_types = MODELSEARCH_STATEMENT_TYPES
        peripheral_name = Name('DRUG')
    elif tool_name == "structsearch":
        statement_types = STRUCTSEARCH_STATEMENT_TYPES
        peripheral_name = Name('MET')

    filtered_statements = []
    for statement in statements:
        if isinstance(statement, statement_types):
            filtered_statements.append(statement)
        elif isinstance(statement, Peripherals) and peripheral_name in statement.modes:
            filtered_statements.append(Peripherals(statement.counts, (peripheral_name,)))
    return tuple(filtered_statements)



import pytest

from pharmpy.model.external.rxode import convert_model


def test_model(testdata, load_model_for_test):
    nmmodel = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = convert_model(nmmodel)
    assert 'rxode2' in model.model_code
    assert 'thetas' in model.model_code
    assert 'omegas' in model.model_code
    assert 'sigmas' in model.model_code


def test_pheno_real(testdata, load_model_for_test):
    nmmodel = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    model = convert_model(nmmodel)
    assert '} else {' in model.model_code


def test_pheno_error_model(testdata, load_model_for_test):
    mmodel = load_model_for_test(testdata / 'nonmem' / 'pheno_real.mod')
    model = convert_model(mmodel)
    assert "Y <- F + SIGMA_1_1*W" in model.model_code


def test_sigma(testdata, load_model_for_test):
    nmmodel = load_model_for_test(testdata / 'nonmem' / 'models' / 'mox2.mod')
    model = convert_model(nmmodel)
    assert "RUV_PROP ~ 0.1" in model.model_code


def test_dataset_modifications(testdata, load_model_for_test):
    nmmodel = load_model_for_test(testdata / 'nonmem' / 'pheno.mod')
    model = convert_model(nmmodel)
    assert "EVID" in model.dataset.columns


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_w_error_model(testdata, load_model_for_test):
    nmmodel = load_model_for_test(testdata / 'nonmem' / 'models' / 'fviii6.mod')
    model = convert_model(nmmodel)
    assert 'W <- sqrt(IPRED**2*THETA_4**2 + THETA_3**2)' in model.model_code


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_omega_same(testdata, load_model_for_test):
    nmmodel = load_model_for_test(testdata / 'nonmem' / 'models' / 'fviii6.mod')
    model = convert_model(nmmodel)
    assert "IOV_CL_7" in model.parameters
    assert "IOV_V_7" in model.parameters



import warnings
from dataclasses import dataclass, replace
from typing import Optional

import pytest

from pharmpy.internals.fs.cwd import chdir
from pharmpy.modeling import set_instantaneous_absorption
from pharmpy.tools import read_results
from pharmpy.workflows import (
    Results,
    Task,
    ToolDatabase,
    Workflow,
    WorkflowBuilder,
    execute_workflow,
    local_dask,
)
from pharmpy.workflows.results import ModelfitResults

# All workflow tests are run by the same xdist test worker
# This is to limit the number of sporadic failures on GHA on Windows
# The failures seem like races in dask distributed because the tmp-dir takes to long to create
# on C: (main drive is D:).


def ignore_scratch_warning():
    warnings.filterwarnings(
        "ignore",
        message=".*creating scratch directories is taking a surprisingly long time",
        category=UserWarning,
    )


@pytest.mark.xdist_group(name="workflow")
def test_execute_workflow_constant(tmp_path):
    a = lambda: 1  # noqa E731
    t1 = Task('t1', a)
    wb = WorkflowBuilder(tasks=[t1], name='test-workflow')
    wf = Workflow(wb)

    with chdir(tmp_path):
        with warnings.catch_warnings():
            ignore_scratch_warning()
            res = execute_workflow(wf)

    assert res == a()


@pytest.mark.xdist_group(name="workflow")
def test_execute_workflow_unary(tmp_path):
    a = lambda: 2  # noqa E731
    f = lambda x: x**2  # noqa E731
    t1 = Task('t1', a)
    t2 = Task('t2', f)
    wb = WorkflowBuilder(tasks=[t1], name='test-workflow')
    wb.add_task(t2, predecessors=[t1])
    wf = Workflow(wb)

    with chdir(tmp_path):
        with warnings.catch_warnings():
            ignore_scratch_warning()
            res = execute_workflow(wf)

    assert res == f(a())


@pytest.mark.xdist_group(name="workflow")
def test_execute_workflow_binary(tmp_path):
    a = lambda: 1  # noqa E731
    b = lambda: 2  # noqa E731
    f = lambda x, y: x + y  # noqa E731
    t1 = Task('t1', a)
    t2 = Task('t2', b)
    t3 = Task('t3', f)
    wb = WorkflowBuilder(tasks=[t1, t2], name='test-workflow')
    wb.add_task(t3, predecessors=[t1, t2])
    wf = Workflow(wb)

    with chdir(tmp_path):
        with warnings.catch_warnings():
            ignore_scratch_warning()
            res = execute_workflow(wf)

    assert res == f(a(), b())


@pytest.mark.xdist_group(name="workflow")
def test_execute_workflow_map_reduce(tmp_path):
    n = 10
    f = lambda x: x**2  # noqa E731
    layer_init = list(map(lambda i: Task(f'x{i}', lambda: i), range(n)))
    layer_map = list(map(lambda i: Task(f'f(x{i})', f), range(n)))
    layer_reduce = [Task('reduce', lambda *y: sum(y))]
    wb = WorkflowBuilder(tasks=layer_init, name='test-workflow')
    wb.insert_workflow(WorkflowBuilder(tasks=layer_map))
    wb.insert_workflow(WorkflowBuilder(tasks=layer_reduce))
    wf = Workflow(wb)

    with chdir(tmp_path):
        with warnings.catch_warnings():
            ignore_scratch_warning()
            res = execute_workflow(wf)

    assert res == sum(map(f, range(n)))


@pytest.mark.xdist_group(name="workflow")
def test_execute_workflow_set_instantaneous_absorption(load_model_for_test, testdata, tmp_path):
    model1 = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan1.mod')
    model2 = load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan2.mod')
    advan1_before = model1.model_code

    t1 = Task('init', lambda x: x, model2)
    t2 = Task('update', set_instantaneous_absorption)
    t3 = Task('postprocess', lambda x: x)
    wb = WorkflowBuilder(tasks=[t1], name='test-workflow')
    wb.insert_workflow(WorkflowBuilder(tasks=[t2]))
    wb.insert_workflow(WorkflowBuilder(tasks=[t3]))
    wf = Workflow(wb)

    with chdir(tmp_path):
        with warnings.catch_warnings():
            ignore_scratch_warning()
            res = execute_workflow(wf)

    assert res.model_code == advan1_before


@pytest.mark.xdist_group(name="workflow")
def test_execute_workflow_fit_mock(load_model_for_test, testdata, tmp_path):
    models = (
        load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan1.mod'),
        load_model_for_test(testdata / 'nonmem' / 'modeling' / 'pheno_advan2.mod'),
    )
    indices = range(len(models))
    ofvs = [(-17 + x) ** 2 - x + 3 for x in indices]

    def fit(ofv, m):
        m = m.replace(modelfit_results=ModelfitResults(ofv=ofv))
        return m

    init = map(lambda i: Task(f'init_{i}', lambda x: x, models[i]), indices)
    process = map(lambda i: Task(f'fit{i}', fit, ofvs[i]), indices)
    wb = WorkflowBuilder(tasks=init, name='test-workflow')
    wb.insert_workflow(WorkflowBuilder(tasks=process))
    gather = Task('gather', lambda *x: x)
    wb.insert_workflow(WorkflowBuilder(tasks=[gather]))
    wf = Workflow(wb)

    with chdir(tmp_path):
        with warnings.catch_warnings():
            ignore_scratch_warning()
            execute_workflow(wf)

    # FIXME: These cannot be updated in place
    # for orig, fitted, ofv in zip(models, res, ofvs):
    #    assert orig.modelfit_results.ofv == ofv
    #    assert fitted.modelfit_results.ofv == ofv
    #    assert orig.modelfit_results == fitted.modelfit_results


@pytest.mark.xdist_group(name="workflow")
def test_execute_workflow_results(tmp_path):
    ofv = 3
    mfr = ModelfitResults(ofv=ofv)

    wb = WorkflowBuilder(tasks=[Task('result', lambda: mfr)], name='test-workflow')
    wf = Workflow(wb)

    with chdir(tmp_path):
        with warnings.catch_warnings():
            ignore_scratch_warning()
            res = execute_workflow(wf)

    assert res.ofv == ofv
    assert not hasattr(res, 'tool_database')


@dataclass(frozen=True)
class MyResults(Results):
    ofv: Optional[float] = None
    tool_database: Optional[ToolDatabase] = None


@pytest.mark.xdist_group(name="workflow")
def test_execute_workflow_results_with_tool_database(tmp_path):
    ofv = 3
    mfr = MyResults(ofv=ofv)

    wb = WorkflowBuilder(tasks=[Task('result', lambda: mfr)], name='test-workflow')
    wf = Workflow(wb)

    with chdir(tmp_path):
        with warnings.catch_warnings():
            ignore_scratch_warning()
            res = execute_workflow(wf)
        assert res.tool_database is not None

    assert res.ofv == ofv


@pytest.mark.xdist_group(name="workflow")
def test_execute_workflow_results_with_report(testdata, tmp_path):
    mfr = replace(read_results(testdata / 'frem' / 'results.json'), tool_database=None)

    wb = WorkflowBuilder(tasks=[Task('result', lambda: mfr)], name='test-workflow')
    wf = Workflow(wb)

    with chdir(tmp_path):
        with warnings.catch_warnings():
            ignore_scratch_warning()
            res = execute_workflow(wf)
        html = res.tool_database.path / 'results.html'
        assert html.is_file()
        assert html.stat().st_size > 500000


@pytest.mark.xdist_group(name="workflow")
def test_local_dispatcher():
    wb = WorkflowBuilder(tasks=[Task('results', lambda x: x, 'input')])
    wf = Workflow(wb)
    res = local_dask.run(wf)
    assert res == 'input'



from difflib import unified_diff
from itertools import islice
from typing import Any, List


def _diff_string_lists(a: List[str], b: List[str]):
    return islice(
        unified_diff(a, b, n=0, lineterm=''),
        2,
        None,
    )


def _diff_texts(a: str, b: str):
    return _diff_string_lists(a.splitlines(), b.splitlines())


def diff(a: Any, b: Any):
    lines = _diff_texts(str(a), str(b))
    joined = '\n'.join(lines)
    return '' if joined == '' else joined + '\n'



from pathlib import Path
from typing import Union

from pharmpy.internals.fs.path import normalize_user_given_path
from pharmpy.internals.fs.tmp import TemporaryDirectory
from pharmpy.workflows.results import Results


def create_report(results: Results, path: Union[Path, str]):
    """Create standard report for results

    The report will be an html created at specified path.

    Parameters
    ----------
    results : Results
        Results for which to create report
    path : Path
        Path to report file
    """
    import pharmpy.reporting.reporting as reporting

    path = normalize_user_given_path(path)

    with TemporaryDirectory() as tmpdirname:
        tmp_path = Path(tmpdirname)
        json_path = tmp_path / 'results.json'
        results.to_json(json_path)
        reporting.generate_report(results.rst_path, json_path, path)



"""
Generic NONMEM option record class.

Assumes 'KEY=VALUE' or 'VALUE' and does not support 'KEY VALUE' in general.
"""

import re
from collections import namedtuple
from typing import Iterable, Optional, Tuple, Union, cast

from pharmpy.internals.parse import AttrToken, AttrTree, NoSuchRuleException
from pharmpy.internals.parse.generic import eval_token

from .record import Record


def _get_key(node: AttrTree) -> str:
    return cast(str, eval_token(node.leaf('KEY')))


def _get_value(node: AttrTree) -> Optional[str]:
    try:
        return cast(str, eval_token(node.leaf('VALUE')))
    except NoSuchRuleException:
        return None


Option = namedtuple('Option', ['key', 'value'])


class OptionRecord(Record):
    def __init__(self, name, raw_name, root):
        if hasattr(self, 'option_defs'):
            self.parsed_options = self.option_defs.parse_ast(root)
        super().__init__(name, raw_name, root)

    @property
    def option_pairs(self):
        """Extract the key-value pairs
        If no value exists set it to None
        Can only handle cases where options are supposed to be unique
        """
        return {_get_key(node): _get_value(node) for node in self.root.subtrees('option')}

    @property
    def all_options(self):
        """Extract all options even if non-unique.
        returns a list of named two-tuples with key and value
        """
        pairs = []
        for node in self.root.subtrees('option'):
            pairs += [Option(_get_key(node), _get_value(node))]
        return pairs

    def get_option(self, name):
        for opt in self.all_options:
            if opt.key[:3] == name[:3]:
                return opt.value
        return None

    def has_option(self, name):
        return name in self.option_pairs

    def get_option_startswith(self, s):
        for opt in self.option_pairs:
            if opt.startswith(s):
                return opt
        return None

    def get_option_lists(self, option):
        """Generator for lists of one option

        For example COMPARTMENT in $MODEL. This handles 'KEY VALUE' syntax.
        """
        next_value = False
        for node in self.root.subtrees('option'):
            value = None
            if next_value:
                value = _get_key(node)
                next_value = False
            elif _get_key(node) == option[: len(_get_key(node))]:
                value = _get_value(node)
                if value is None:
                    next_value = True
            if value is not None:
                if value[0] == '(' and value[-1] == ')':
                    yield re.split(r'\s+|,', value[1:-1])
                else:
                    yield [value]

    def set_option(self, key: str, new_value: str):
        """Set the value of an option

        If option already exists replaces its value
        appends option at the end if it does not exist
        does not handle abbreviations yet
        """
        # If already exists update value
        last_option = None
        new_children = []
        it = iter(self.root.children)
        for node in it:
            if not isinstance(node, AttrTree) or node.rule != 'option':
                new_children.append(node)
                continue

            if _get_key(node) == key:
                new_children.append(node.replace_first(AttrToken('VALUE', new_value)))
                new_children.extend(it)
                newroot = AttrTree(self.root.rule, tuple(new_children))
                return self.__class__(self.name, self.raw_name, newroot)

            new_children.append(node)

            last_option = node

        ws_token = AttrToken('WS', ' ')
        option_node = self._create_option(key, new_value)
        # If no other options add first else add just after last option
        if last_option is None:
            newroot = AttrTree(
                self.root.rule,
                (
                    ws_token,
                    option_node,
                )
                + self.root.children,
            )
        else:
            new_children = []
            for node in self.root.children:
                new_children.append(node)
                if node is last_option:
                    new_children += [ws_token, option_node]
            newroot = AttrTree(self.root.rule, tuple(new_children))
        return self.__class__(self.name, self.raw_name, newroot)

    def _create_option(self, key: str, value: Optional[str] = None):
        if value is None:
            node = AttrTree.create('option', [{'KEY': key}])
        else:
            node = AttrTree.create('option', [{'KEY': key}, {'EQUAL': '='}, {'VALUE': value}])
        return node

    def prepend_option(self, key: str, value: Optional[str] = None):
        """Prepend option"""
        node = self._create_option(key, value)
        new_root = self._prepend_option_node(node)
        return self.__class__(self.name, self.raw_name, new_root)

    def _prepend_option_node(self, node):
        """Add a new option as first option"""
        ws_token = AttrToken('WS', ' ')
        new = (node, ws_token)
        return AttrTree(self.root.rule, self.root.children[:1] + new + self.root.children[1:])

    def append_option(self, key: str, value: Optional[str] = None):
        """Append option as last option

        Method applicable to option records with no special grammar
        """
        node = self._create_option(key, value)
        newrec = self.append_option_node(node)
        return newrec

    def _append_option_args(self) -> Tuple[int, int, AttrToken]:
        children = self.root.children
        n = len(children)
        # NOTE: Pop trailing whitespace if any
        j = n - 1 if children[-1].rule == 'WS' else n
        for i, child in zip(reversed(range(n)), reversed(children)):
            rule = child.rule
            if rule == 'option':
                return (i + 1, j, AttrToken('WS', ' '))
            elif rule not in ('WS', 'NEWLINE'):
                return (i + 1, j, AttrToken('NEWLINE', '\n'))

        return (0, j, AttrToken('WS', ' '))

    def append_option_node(self, node):
        """Add a new option as last option"""
        i, j, sep = self._append_option_args()
        children = self.root.children
        newroot = AttrTree(self.root.rule, children[:i] + (sep, node) + children[i:j])
        return self.__class__(self.name, self.raw_name, newroot)

    def replace_option(self, old, new):
        """Replace an option"""

        def _fn(node: Union[AttrTree, AttrToken]):
            if isinstance(node, AttrTree) and node.rule == 'option':
                if node.find('KEY') is not None:
                    if eval_token(node.leaf('KEY')) == old:
                        return node.replace_first(AttrToken('KEY', new))
                elif node.find('VALUE') is not None:
                    if eval_token(node.leaf('VALUE')) == old:
                        return node.replace_first(AttrToken('VALUE', new))

            return node

        newroot = self.root.map(_fn)
        return self.__class__(self.name, self.raw_name, newroot)

    def remove_option(self, key):
        """Remove all options key"""
        new_children = []
        for node in self.root.children:
            if isinstance(node, AttrTree) and node.rule == 'option' and key == _get_key(node):
                if new_children[-1].rule == 'WS':
                    new_children.pop()
            else:
                new_children.append(node)
        newroot = AttrTree(self.root.rule, tuple(new_children))
        return self.__class__(self.name, self.raw_name, newroot)

    def remove_nth_option(self, key, n):
        """Remove the nth option key"""
        new_children = []
        i = 0
        for node in self.root.children:
            if isinstance(node, AttrTree) and node.rule == 'option':
                curkey = _get_key(node)
                if key[: len(curkey)] == curkey and i == n:
                    if new_children[-1].rule == 'WS':
                        new_children.pop()
                else:
                    new_children.append(node)
                if key[: len(curkey)] == curkey:
                    i += 1
            else:
                new_children.append(node)
        newroot = AttrTree(self.root.rule, tuple(new_children))
        return self.__class__(self.name, self.raw_name, newroot)

    def add_suboption_for_nth(self, key, n, suboption):
        """Adds a suboption to the nth option key"""
        i = 0
        new_children = []
        it = iter(self.root.children)
        for node in it:
            if isinstance(node, AttrTree) and node.rule == 'option':
                curkey = _get_key(node)
                if key[: len(curkey)] == curkey:
                    if i == n:
                        s = node.leaf('VALUE').value
                        if s.startswith('('):
                            s = f'{s[:-1]} {suboption})'
                        else:
                            s = f'({s} {suboption})'
                        new_children.append(node.replace_first(AttrToken('VALUE', s)))
                        new_children.extend(it)
                        newroot = AttrTree(self.root.rule, tuple(new_children))
                        return self.__class__(self.name, self.raw_name, newroot)
                    i += 1

            new_children.append(node)

    def remove_suboption_for_all(self, key, suboption):
        """Remove subtoption from all options key"""
        new_children = []
        it = iter(self.root.children)
        for node in it:
            if isinstance(node, AttrTree) and node.rule == 'option':
                curkey = _get_key(node)
                if key[: len(curkey)] == curkey:
                    s = node.leaf('VALUE').value
                    if s.startswith('('):
                        subopts = [
                            subopt
                            for subopt in s[1:-1].split()
                            if suboption[: len(subopt)] != subopt
                        ]
                        s = '(' + ' '.join(subopts) + ')'
                        new_children.append(node.replace_first(AttrToken('VALUE', s)))
                        continue

            new_children.append(node)

        newroot = AttrTree(self.root.rule, tuple(new_children))
        return self.__class__(self.name, self.raw_name, newroot)

    def remove_option_startswith(self, start):
        """Remove all options that startswith"""
        rec = self
        for key in self.option_pairs:
            if key.startswith(start):
                rec = rec.remove_option(key)
        return rec

    @staticmethod
    def match_option(options: Iterable[str], query: str):
        """Match a given option to any from a set of valid options

        NONMEM allows matching down to three letters as long as
        there are no ambiguities.

        return the canonical form of the matched option or None for no match
        """

        min_prefix_len = 3

        if len(query) < min_prefix_len:
            # NOTE: This keeps the original implementation's behavior but maybe
            # this could be changed?
            return None

        i: int = min_prefix_len

        candidates = [option for option in options if option[:i] == query[:i]]

        while len(candidates) >= 2 and i < len(query):
            candidates = [
                option for option in candidates if i < len(option) and option[i] == query[i]
            ]
            i += 1

        return candidates[0] if len(candidates) == 1 else None


class Opts:
    def __init__(self, *args):
        self.options = args
        self.find_abbreviations()

    def __getitem__(self, name):
        for opt in self.options:
            if opt.name == name:
                return opt
        raise KeyError(f"No option named {name}")

    def find_abbreviations(self):
        accepted = set()
        done = set()

        # Handle options not allowed to be abbreviated
        for opt in self.options:
            if not isinstance(opt, WildOpt) and opt.noabbrev:
                # Currenty only single options
                opt.abbreviations = [opt.name]
                accepted.add(opt.name)
                done.add(opt.name)

        # Find longest length
        longest = 0
        for opt in self.options:
            if isinstance(opt, WildOpt):
                continue
            if len(opt.name) > longest:
                longest = len(opt.name)

        curlength = longest
        while True:
            count = {}
            optdict = {}
            for opt in self.options:
                if isinstance(opt, WildOpt):
                    continue
                name = opt.name
                if name not in done and len(name) >= curlength:
                    abbrev = name[0:curlength]
                    if abbrev not in count:
                        count[abbrev] = 1
                        optdict[abbrev] = opt
                    else:
                        count[abbrev] += 1

            if len(count) == 0:
                continue

            for name, n in count.items():
                if n == 1 and name not in accepted:
                    optdict[name].abbreviations.append(name)
                    accepted.add(name)
                else:
                    done.add(name)

            if curlength == 3:
                break
            curlength -= 1

    def parse_ast(self, tree, nonoptions=None, netas=None):
        # Return a list of tuples of canonical option name, value (or None for no value)
        # Nonoptions will not be matched against options but kept as WildOpts
        opt_nodes = list(tree.subtrees('option'))
        found_options = set()
        i = 0
        parsed = []

        def add_option(opt, name, value):
            # * Same MxOpt or SimpleOpt multiple times is ignored
            # * Same opt taking argument multiple times is illegal
            # * Another option in same mx group is illegal
            if opt in found_options:
                if opt.need_value:
                    raise ValueError(f"Option {name} has already been specified.")
                else:
                    return
            elif isinstance(opt, MxOpt):
                for fopt in found_options:
                    if isinstance(fopt, MxOpt) and fopt.group == opt.group and fopt.name != name:
                        raise ValueError(
                            f"Option {fopt.name} in same group as {name} has already been specified."
                        )
            parsed.append((opt, name, value))
            found_options.add(opt)

        while i < len(opt_nodes):
            node = opt_nodes[i]
            key = _get_key(node)
            # FIXME: This is special for $TABLE. Either have special case for this or have an Opt for it.
            m = re.match(r'ETAS\(?', key)
            if m:
                # join all keys that belong with the ETAS
                etas = key
                while ')' not in key:
                    i += 1
                    node = opt_nodes[i]
                    key = _get_key(node)
                    etas += key
                etas = etas[5:-1].replace(" ", "").replace("\t", "")
                if ',' in etas:
                    a = etas.split(',')
                    for n in a:
                        parsed.append((WildOpt(), f'ETA{n}', None))
                else:
                    m = re.match(r'(?P<start>\d+)(TO|:)(?P<end>\d+|LAST)(BY(?P<by>-?\d+))?', etas)
                    if m:
                        by = 1 if m.group('by') is None else int(m.group('by'))
                        start = int(m.group('start'))
                        end = m.group('end')
                        if end == 'LAST':
                            end = netas
                        else:
                            end = int(end)
                        if start > end:
                            end -= 1
                            if by > 0:
                                by = -by
                        else:
                            if by < 0:
                                start, end = end, start
                                end -= 1
                            else:
                                end += 1
                        for n in range(start, end, by):
                            parsed.append((WildOpt(), f'ETA{n}', None))
                i += 1
                continue

            # FIXME: What happens if key=value for nonoption?
            if key in nonoptions:
                parsed.append((WildOpt(), key, None))
                i += 1
                continue
            value = _get_value(node)
            for opt in self.options:
                if opt.match(key):
                    if opt.need_value is True:
                        if value is None:
                            i += 1
                            if i < len(opt_nodes):
                                next_node = opt_nodes[i]
                                next_key = _get_key(next_node)
                                next_value = _get_value(next_node)
                                if next_value is not None:
                                    raise ValueError(f"Unexpected value for {opt.name}")
                                value = next_key
                            else:
                                raise ValueError(f"No value for option {opt.name}")
                        converted = opt.convert_value(value)
                        if converted is not None:
                            add_option(opt, opt.name, converted)
                        else:
                            raise ValueError(f"Bad value {value} for option {opt.name}")

                    elif opt.need_value is False:
                        if value is not None:
                            raise ValueError(f"Unexpected value for {opt.name}")
                        add_option(opt, opt.name, value)
                    else:  # value optional
                        parsed.append((opt, key, value))
                    break
            i += 1
        return parsed


class Opt:
    def __init__(self, noabbrev=False):
        self.abbreviations = []
        self.noabbrev = noabbrev

    def __eq__(self, other):
        return self.name == other.name

    def __hash__(self):
        return hash(self.name)

    def match(self, s):
        return s in self.abbreviations


class MxOpt(Opt):
    need_value = False

    def __init__(self, name, group, default=False, **kwargs):
        self.name = name
        self.group = group
        self.default = default
        super().__init__(**kwargs)


class StrOpt(Opt):
    need_value = True

    def __init__(self, name, default=None, **kwargs):
        self.name = name
        self.default = default
        super().__init__(**kwargs)

    def convert_value(self, value):
        return value


class IntOpt(Opt):
    need_value = True

    def __init__(self, name, default=None, **kwargs):
        self.name = name
        self.default = default
        super().__init__(**kwargs)

    def convert_value(self, value):
        try:
            n = int(value)
        except ValueError:
            return None
        return n


class SimpleOpt(Opt):
    need_value = False

    def __init__(self, name, **kwargs):
        self.name = name
        super().__init__(**kwargs)


class EnumOpt(Opt):
    need_value = True

    def __init__(self, name, allowed, default=None, **kwargs):
        self.name = name
        self.allowed = allowed
        if not (default is None or default in allowed):
            raise ValueError(f"Default value {default} must be in allowed: {allowed}")
        super().__init__(**kwargs)

    def convert_value(self, value):
        enum = value.upper()
        if enum in self.allowed:
            return enum
        else:
            return None


class WildOpt:
    need_value = None

    def match(self, s):
        return True

    def __eq__(self, other):
        return isinstance(other, WildOpt)



import pytest
import sympy

from pharmpy.model import ModelSyntaxError, Parameter, Parameters


def S(x):
    return sympy.Symbol(x)


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf,results',
    [
        ('$OMEGA 1', [([None], [1.0], False, False)]),
        ('$OMEGA   0.123 \n\n', [([None], [0.123], False, False)]),
        ('$OMEGA   (0 FIX) ; CL', [(['CL'], [0.0], True, False)]),
        (
            '$OMEGA DIAG(2) 1 2 FIX',
            [
                ([None], [1.0], False, False),
                ([None], [2.0], True, False),
            ],
        ),
        (
            '$OMEGA 1 2 3',
            [
                ([None], [1.0], False, False),
                ([None], [2.0], False, False),
                ([None], [3.0], False, False),
            ],
        ),
        ('$OMEGA 0.15 ;CL', [(['CL'], [0.15], False, False)]),
        (
            '$OMEGA 1 \n2 ; S   \n  3 ',
            [
                ([None], [1.0], False, False),
                (['S'], [2.0], False, False),
                ([None], [3.0], False, False),
            ],
        ),
        ('$OMEGA 2 SD', [([None], [4.0], False, False)]),
        ('$OMEGA ;CO\n (VAR 3)', [([None], [3.0], False, False)]),
        (
            '$OMEGA (1)x2',
            [
                ([None], [1.0], False, False),
                ([None], [1.0], False, False),
            ],
        ),
        (
            '$OMEGA BLOCK(2) 1 0.5 2',
            [
                ([None, None, None], [1.0, 0.5, 2.0], False, False),
            ],
        ),
        ('$OMEGA BLOCK(2) SAME', [(None, None, None, True)]),
        ('$OMEGA BLOCK SAME(3)', [(None, None, None, True)]),
        (
            '$OMEGA BLOCK(2) 1 (0.1)x2',
            [
                ([None, None, None], [1.0, 0.1, 0.1], False, False),
            ],
        ),
        (
            '$OMEGA BLOCK(2) CHOLESKY 0.8 -0.3 0.7',
            [
                ([None, None, None], [pytest.approx(0.64), -0.24, 0.58], False, False),
            ],
        ),
        (
            '$OMEGA BLOCK(2) SD 0.8 -0.394 0.762 CORR',
            [
                (
                    [None, None, None],
                    [pytest.approx(0.64), pytest.approx(-0.2401824), pytest.approx(0.580644)],
                    False,
                    False,
                ),
            ],
        ),
        (
            '$OMEGA BLOCK(1)   1.5',
            [
                ([None], [1.5], False, False),
            ],
        ),
        (
            '$OMEGA  0.0258583  ;      V2\n'
            ';$OMEGA BLOCK(1) 0.0075 FIX    ;.02 ; IOC\n'
            ';$OMEGA BLOCK(1) SAME\n',
            [
                (['V2'], [0.0258583], False, False),
            ],
        ),
        ('$OMEGA 1 ; IVCL', [(['IVCL'], [1.0], False, False)]),
        (
            '$OMEGA DIAG(2) 1 ; V1 df\n 2 FIX ; VA2 __12\n',
            [
                (['V1'], [1.0], False, False),
                (['VA2'], [2.0], True, False),
            ],
        ),
        (
            '$OMEGA BLOCK(2) 1 ;IV1\n 2 ;CORR\n 3 ;IV2',
            [
                (['IV1', 'CORR', 'IV2'], [1.0, 2.0, 3.0], False, False),
            ],
        ),
    ],
)
def test_parse(parser, buf, results):
    recs = parser.parse(buf)
    rec = recs.records[0]
    assert rec.parse() == results


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf, exc_msg',
    [
        ('$OMEGA 0 ', 'If initial estimate for OMEGA is'),
        ('$OMEGA DIAG(1) 1 SD VARIANCE', 'Initial estimate for OMEGA cannot'),
        ('$OMEGA SD BLOCK(2) 0.1 0.001 0.1 STANDARD', 'Cannot specify either option'),
        ('$OMEGA CHOLESKY BLOCK(2) 0.1 VAR 0.001 \n  0.1 ', 'Cannot specify either option'),
        ('$OMEGA BLOCK(3) 0.1 \n  0.1', 'Wrong number of inits in BLOCK'),
    ],
)
def test_errors(parser, buf, exc_msg):
    recs = parser.parse(buf)
    rec = recs.records[0]
    with pytest.raises(ModelSyntaxError, match=exc_msg):
        pset, _, _ = rec.parse()


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf, params, results',
    [
        ('$OMEGA 1', [Parameter('OMEGA_1_1', init=2.0, lower=0.0)], '$OMEGA 2'),
        ('$OMEGA 1 SD', [Parameter('OMEGA_1_1', init=4.0, lower=0.0)], '$OMEGA 2 SD'),
        (
            '$OMEGA (1)x3\n;FTOL',
            [
                Parameter('OMEGA_1_1', init=4.0, lower=0.0),
                Parameter('OMEGA_2_2', init=4.0, lower=0.0),
                Parameter('OMEGA_3_3', init=4.0, lower=0.0),
            ],
            '$OMEGA (4)x3\n;FTOL',
        ),
        (
            '$OMEGA (1)x2 2',
            [
                Parameter('OMEGA_1_1', init=1.0, lower=0.0),
                Parameter('OMEGA_2_2', init=2.0, lower=0.0),
                Parameter('OMEGA_3_3', init=0.5, lower=0.0),
            ],
            '$OMEGA (1) (2) 0.5',
        ),
        (
            '$OMEGA DIAG(2) (1 VAR) (SD 2)',
            [
                Parameter('OMEGA_1_1', init=1.5, lower=0.0),
                Parameter('OMEGA_2_2', init=16.0, lower=0.0),
            ],
            '$OMEGA DIAG(2) (1.5 VAR) (SD 4)',
        ),
        (
            '$OMEGA BLOCK(2) 1 2 4',
            [
                Parameter('OMEGA_1_1', init=7.0, lower=0.0),
                Parameter('OMEGA_2_1', init=0.5),
                Parameter('OMEGA_2_2', init=8.0, lower=0.0),
            ],
            '$OMEGA BLOCK(2) 7 0.5 8',
        ),
        (
            '$OMEGA BLOCK(2)\n SD 1 0.5 ;COM \n 1\n',
            [
                Parameter('OMEGA_1_1', init=4.0, lower=0.0),
                Parameter('OMEGA_2_1', init=0.25),
                Parameter('OMEGA_2_2', init=9.0, lower=0.0),
            ],
            '$OMEGA BLOCK(2)\n SD 2 0.25 ;COM \n 3\n',
        ),
        (
            '$OMEGA CORR BLOCK(2)  1 0.5 1\n',
            [
                Parameter('OMEGA_1_1', init=4.0, lower=0.0),
                Parameter('OMEGA_2_1', init=1.0),
                Parameter('OMEGA_2_2', init=4.0, lower=0.0),
            ],
            '$OMEGA CORR BLOCK(2)  4 0.25 4\n',
        ),
        (
            '$OMEGA CORR BLOCK(2)  1 0.5 SD 1\n',
            [
                Parameter('OMEGA_1_1', init=4.0, lower=0.0),
                Parameter('OMEGA_2_1', init=1.0),
                Parameter('OMEGA_2_2', init=4.0, lower=0.0),
            ],
            '$OMEGA CORR BLOCK(2)  2 0.25 SD 2\n',
        ),
        (
            '$OMEGA BLOCK(2) 1 0.1 1\n CHOLESKY',
            [
                Parameter('OMEGA_1_1', init=0.64, lower=0.0),
                Parameter('OMEGA_2_1', init=-0.24),
                Parameter('OMEGA_2_2', init=0.58, lower=0.0),
            ],
            '$OMEGA BLOCK(2) 0.8 -0.3 0.7\n CHOLESKY',
        ),
        (
            '$OMEGA BLOCK(2) (1)x3\n',
            [
                Parameter('OMEGA_1_1', init=0.64, lower=0.0),
                Parameter('OMEGA_2_1', init=-0.24),
                Parameter('OMEGA_2_2', init=0.58, lower=0.0),
            ],
            '$OMEGA BLOCK(2) 0.64 -0.24 0.58\n',
        ),
        (
            '$OMEGA BLOCK(3) 1 ;CL\n0.1 1; V\n0.1 0.1 1; KA',
            [
                Parameter('OMEGA_1_1', init=1.0, lower=0.0),
                Parameter('OMEGA_2_1', init=0.2),
                Parameter('OMEGA_2_2', init=3.0, lower=0.0),
                Parameter('OMEGA_3_1', init=0.4),
                Parameter('OMEGA_3_2', init=0.5),
                Parameter('OMEGA_3_3', init=6.0, lower=0.0),
            ],
            '$OMEGA BLOCK(3) 1 ;CL\n0.2 3; V\n0.4 0.5 6; KA',
        ),
        ('$OMEGA BLOCK(2) SAME', [], '$OMEGA BLOCK(2) SAME'),
        ('$OMEGA BLOCK SAME', [], '$OMEGA BLOCK SAME'),
        (
            '$OMEGA 2 FIX 4 (FIX 6)',
            [
                Parameter('OMEGA_1_1', init=2.0, lower=0.0, fix=False),
                Parameter('OMEGA_2_2', init=4.0, lower=0.0),
                Parameter('OMEGA_3_3', init=6.0, lower=0.0, fix=True),
            ],
            '$OMEGA 2 4 (FIX 6)',
        ),
        (
            '$OMEGA 2 4 6 ;STRAML',
            [
                Parameter('OMEGA_1_1', init=2.0, lower=0.0, fix=True),
                Parameter('OMEGA_2_2', init=4.0, lower=0.0),
                Parameter('OMEGA_3_3', init=6.0, lower=0.0),
            ],
            '$OMEGA 2 FIX 4 6 ;STRAML',
        ),
        (
            '$OMEGA 2 FIX 4 6 ;STRAML',
            [
                Parameter('OMEGA_1_1', init=2.0, lower=0.0, fix=False),
                Parameter('OMEGA_2_2', init=4.0, lower=0.0, fix=True),
                Parameter('OMEGA_3_3', init=23.0, lower=0.0, fix=True),
            ],
            '$OMEGA 2 4 FIX 23 FIX ;STRAML',
        ),
        (
            '$OMEGA BLOCK(2) 1 .1 2 ;CLERMT',
            [
                Parameter('OMEGA_1_1', init=1.0, lower=0.0, fix=True),
                Parameter('OMEGA_2_1', init=0.1, fix=True),
                Parameter('OMEGA_2_2', init=2.0, lower=0.0, fix=True),
            ],
            '$OMEGA BLOCK(2) FIX 1 .1 2 ;CLERMT',
        ),
        (
            '$OMEGA BLOCK(2) 1 .1 FIX 1 ;CLERMT',
            [
                Parameter('OMEGA_1_1', init=1.0, lower=0.0, fix=False),
                Parameter('OMEGA_2_1', init=0.1, fix=False),
                Parameter('OMEGA_2_2', init=1.0, lower=0.0, fix=False),
            ],
            '$OMEGA BLOCK(2) 1 .1 1 ;CLERMT',
        ),
    ],
)
def test_update(parser, buf, params, results):
    rec = parser.parse(buf).records[0]
    pset = Parameters(params)
    newrec = rec.update(pset)
    assert str(newrec) == results


@pytest.mark.usefixtures('parser')
@pytest.mark.parametrize(
    'buf, params',
    [
        (
            '$OMEGA BLOCK(2) 1 .1 2 ;CLERMT',
            [
                Parameter('OMEGA_1_1', init=1.0, lower=0.0, fix=True),
                Parameter('OMEGA_2_1', init=0.1),
                Parameter('OMEGA_2_2', init=2.0, lower=0.0, fix=True),
            ],
        ),
    ],
)
def test_update_error(parser, buf, params):
    rec = parser.parse(buf).records[0]
    pset = Parameters(params)
    with pytest.raises(ValueError):
        rec.update(pset)


@pytest.mark.parametrize(
    'buf,remove,result',
    [
        (
            '$OMEGA BLOCK(3) 1 ;CL\n0.1 1; V\n0.1 0.1 1; KA',
            [(0, 0)],
            '$OMEGA BLOCK(2)\n1.0\n0.1 1.0\n',
        ),
        (
            '$OMEGA BLOCK(4) 1 ;CL\n0.2 3; V\n0.4 0.5 6; KA\n7 8 9 10\n',
            [(0, 1), (0, 3)],
            '$OMEGA BLOCK(2)\n1.0\n0.4 6.0\n',
        ),
        ('$OMEGA 1 2 3\n', [(1, 0)], '$OMEGA 1 3\n'),
        ('$OMEGA BLOCK(2) 1 2 3 FIX\n', [(0, 0)], '$OMEGA BLOCK(1) FIX\n3.0\n'),
    ],
)
def test_remove(parser, buf, remove, result):
    rec = parser.parse(buf).records[0]
    newrec = rec.remove(remove)
    assert str(newrec) == result



from typing import Any, Mapping, Tuple

from pharmpy.internals.fs.path import normalize_user_given_path


def split_common_options(d) -> Tuple[Mapping[str, Any], Mapping[str, Any]]:
    """Split the dict into common options and other options

    Parameters
    ----------
    d : dict
        Dictionary of all options

    Returns
    -------
    Tuple of common options and other option dictionaries
    """
    execute_options = ['path', 'resume']
    common_options = {}
    other_options = {}
    for key, value in d.items():
        if key in execute_options:
            if key == 'path':
                if value is not None:
                    value = normalize_user_given_path(value)
            common_options[key] = value
        else:
            other_options[key] = value
    return common_options, other_options



from pharmpy.workflows import Task


def test_create_tasks():
    t1 = Task('t1', 'func', 'input')
    assert t1.task_input == ('input',)
    t2 = Task('t2', 'func', 1)
    assert t2.task_input == (1,)
    t3 = Task('t3', 'func', 1, 2, 3)
    assert t3.task_input == (1, 2, 3)
    t4 = Task.create('t4', 'func', [1, 2, 3])
    assert t4.task_input == ([1, 2, 3],)
    t5 = Task.create('t5', 'func')
    assert t5.task_input == ()


def test_replace():
    t1 = Task('t1', 'func', 'input')
    t2 = t1.replace(name='new')
    assert t1.name != t2.name
